'''
Created on Jun 20, 2020

@author: petergedeck
'''
from collections import namedtuple
from enum import Enum
import numbers

from scipy import stats
from scipy.special import gammaln

import numpy as np
import pandas as pd


GroupMeans = namedtuple('GroupMeans', 'statistics,center')


class QCCStatistics:

    def __init__(self):
        self.default = None
        self.statistics = {}
        self.register(Xbar_statistic, default=True)
        self.register(Xbar_one_statistic)
        self.register(R_statistic)
        self.register(S_statistic)
        self.register(P_statistic)
        self.register(NP_statistic)

    def get(self, name):
        if name is None and self.default is None:
            raise ValueError('you need give a name as default is not defined')
        if name.lower() not in self.statistics:
            raise ValueError(f'unknown statistic {name}')
        return self.statistics[name.lower()]()

    def register(self, statistic, default=False):
        name = statistic.qcc_type
        if name.lower() in self.statistics:
            raise ValueError(f'duplicate definition of statistics {name}')
        if default and self.default is not None:
            raise ValueError(f'duplicate definition of default statistics {name} and {default}')
        name = name.lower()
        if default:
            self.default = name
        self.statistics[name] = statistic

    def __iter__(self):
        if self.default is not None:
            yield self.statistics[self.default]
        for statistic in self.statistics.values():
            if statistic.qcc_type.lower() == self.default:
                continue
            yield statistic


class SD_estimator(Enum):
    uwave_r = 'UWAVE-R'  # UnWeighted AVErage of subgroup estimates based on subgroup Ranges
    uwave_sd = 'UWAVE-SD'  # UnWeighted AVErage of subgroup estimates based on subgroup Standard Deviations
    mvlue_r = 'MVLUE-R'  # Minimum Variance Linear Unbiased Estimator computed as a weighted average of subgroups estimates based on subgroup Ranges
    mvlue_sd = 'MVLUE-SD'  # Minimum Variance Linear Unbiased Estimator computed as a weighted average of subgroup estimates based on subgroup Standard Deviations
    rmsdf = 'RMSDF'  # Root-Mean-Square estimator computed as a weighted average of subgroup estimates based on subgroup Standard Deviations
    mr = 'MR'
    sd = 'SD'

    @classmethod
    def get(cls, name, default):
        if name is None:
            name = default
        if isinstance(name, cls):
            return name
        if not isinstance(name, str):
            raise ValueError(f'name of estimator must be string or SD_estimator, found: {name}')
        name = name.upper()
        for estimator in cls:
            if estimator.value == name:
                return estimator
        raise ValueError(f'unknown SD estimator {name}')


# exp.R.unscaled a vector specifying, for each sample size, the expected value of the relative range
# (i.e. R/σ) for a normal distribution. This appears as d2 on most tables containing factors for
# the construction of control charts.
_exp_R_unscaled = [np.NaN, np.NaN, 1.128, 1.693, 2.059, 2.326, 2.534, 2.704, 2.847, 2.970, 3.078, 3.173,
                   3.258, 3.336, 3.407, 3.472, 3.532, 3.588, 3.640, 3.689, 3.735, 3.778, 3.819, 3.858, 3.895, 3.931]
# se.R.unscaled a vector specifying, for each sample size, the standard error of the relative range
# (i.e. R/σ) for a normal distribution. This appears as d3 on most tables containing factors for
# the construction of control charts.
_se_R_unscaled = [np.NaN, np.NaN, 0.8525033, 0.8883697, 0.8798108, 0.8640855, 0.8480442, 0.8332108, 0.8198378, 0.8078413, 0.7970584, 0.7873230, 0.7784873, 0.7704257, 0.7630330, 0.7562217, 0.7499188, 0.7440627, 0.7386021, 0.7334929, 0.7286980, 0.7241851, 0.7199267, 0.7158987, 0.7120802,
                  0.7084528, 0.7050004, 0.7017086, 0.6985648, 0.6955576, 0.6926770, 0.6899137, 0.6872596, 0.6847074, 0.6822502, 0.6798821, 0.6775973, 0.6753910, 0.6732584, 0.6711952, 0.6691976, 0.6672619, 0.6653848, 0.6635632, 0.6617943, 0.6600754, 0.6584041, 0.6567780, 0.6551950, 0.6536532, 0.6521506]


class Base_statistic:
    @staticmethod
    def getSizes(data):
        return [int(n) for n in (~np.isnan(data)).sum(1)]

    def stats(self, data, sizes=None):
        raise NotImplementedError()

    def sd(self, data, std_dev, sizes=None):
        raise NotImplementedError()

    def limits(self, center, std_dev, sizes, conf):
        raise NotImplementedError()


class Xbar_statistic(Base_statistic):
    """ Statistics used in computing and drawing a Shewhart xbar chart """
    qcc_type = 'xbar'
    description = ('mean', 'means of a continuous process variable')

    def stats(self, data, sizes=None):
        return GroupMeans(np.nanmean(data, axis=1), np.nanmean(data))

    def sd(self, data, std_dev=None, sizes=None):
        if isinstance(std_dev, numbers.Number):
            return std_dev

        if sizes is None:
            sizes = self.getSizes(data)

        # set default method if std_dev is None
        if std_dev is None:
            std_dev = SD_estimator.uwave_r
            if max(sizes) > 25:
                std_dev = SD_estimator.rmsdf

        if std_dev == SD_estimator.uwave_r:
            r = np.nanmax(data, axis=1) - np.nanmin(data, axis=1)
            d2 = [_exp_R_unscaled[size] for size in sizes]
            return np.mean(r / d2)
        elif std_dev == SD_estimator.uwave_sd:
            s = np.nanstd(data, axis=1, ddof=1)
            c4 = np.array([qcc_c4(size) for size in sizes])
            return np.mean(s / c4)
        elif std_dev == SD_estimator.mvlue_r:
            r = np.nanmax(data, axis=1) - np.nanmin(data, axis=1)
            d2 = np.array([_exp_R_unscaled[size] for size in sizes])
            d3 = np.array([_se_R_unscaled[size] for size in sizes])
            w = (d2 / d3) ** 2
            return np.sum(w * r / d2) / np.sum(w)
        elif std_dev == SD_estimator.mvlue_sd:
            s = np.nanstd(data, axis=1, ddof=1)
            c4 = np.array([qcc_c4(size) for size in sizes])
            w = c4 ** 2 / (1 - c4**2)
            return np.sum(w * s / c4) / np.sum(w)
        elif std_dev == SD_estimator.rmsdf:
            s = np.nanstd(data, axis=1, ddof=1)
            w = np.array([size - 1 for size in sizes])
            return np.sqrt(np.sum(w * s**2) / np.sum(w)) / qcc_c4(np.sum(w) + 1)

        else:
            raise NotImplementedError()

    def limits(self, center, std_dev, sizes, conf):
        if conf < 0:
            raise ValueError(f'invalid conf argument {conf}')

        if all(size == sizes[0] for size in sizes):
            sizes = [sizes[0]]

        se_stats = std_dev / np.sqrt(sizes)
        if conf >= 1:
            lcl = center - conf * se_stats
            ucl = center + conf * se_stats
        else:
            sigmas = stats.norm.ppf(1 - (1 - conf) / 2)
            lcl = center - sigmas * se_stats
            ucl = center + sigmas * se_stats
        return pd.DataFrame({'LCL': lcl, 'UCL': ucl})


class Xbar_one_statistic(Base_statistic):
    """ Statistics used in xbar-one charts """
    qcc_type = 'xbarone'
    description = ('mean', 'one-at-time data of a continuous process variable')

    @staticmethod
    def getSizes(data):
        pass  # not required
        return [1 for _ in data]

    def stats(self, data, sizes=None):
        if isinstance(data, (pd.Series, pd.DataFrame)):
            data = data.values
        return GroupMeans(np.array(data).flatten(), np.mean(data))

    def sd(self, data, std_dev=None, sizes=None, k=2):
        if isinstance(std_dev, numbers.Number):
            return std_dev

        if isinstance(data, (pd.Series, pd.DataFrame)):
            data = data.values

        # set default method if std_dev is None
        std_dev = SD_estimator.get(std_dev, SD_estimator.mr)
        std_dev = std_dev or SD_estimator.mr
        if isinstance(std_dev, str):
            std_dev = std_dev.upper()

        if std_dev == SD_estimator.mr:
            d = 0
            for i in range(len(data) - (k - 1)):
                group = data[i:i + k]
                d += np.max(group) - np.min(group)
            return d / ((len(data) - k + 1) * _exp_R_unscaled[k])
        elif std_dev == SD_estimator.sd:
            return np.std(data, ddof=1) / qcc_c4(len(data))
        else:
            raise NotImplementedError(f'estimator {std_dev}')

    def limits(self, center, std_dev, sizes, conf):
        if conf < 0:
            raise ValueError(f'invalid conf argument {conf}')
        se_stats = std_dev
        if conf >= 1:
            lcl = center - conf * se_stats
            ucl = center + conf * se_stats
        else:
            sigmas = stats.norm.ppf(1 - (1 - conf) / 2)
            lcl = center - sigmas * se_stats
            ucl = center + sigmas * se_stats
        if isinstance(lcl, numbers.Number):
            lcl = [lcl]
            ucl = [ucl]
        return pd.DataFrame({'LCL': lcl, 'UCL': ucl})


class R_statistic(Base_statistic):
    """ Statistics used in computing and drawing a Shewhart R chart """
    qcc_type = 'R'
    description = ('range', ' ranges of a continuous process variable')

    def stats(self, data, sizes=None):
        if sizes is None:
            sizes = self.getSizes(data)
        by_group = np.nanmax(data, axis=1) - np.nanmin(data, axis=1)
        return GroupMeans(by_group, sum(sizes * by_group) / sum(sizes))

    def sd(self, data, std_dev=None, sizes=None):
        if isinstance(std_dev, numbers.Number):
            return std_dev

        if std_dev is None:
            std_dev = SD_estimator.uwave_r

        if std_dev not in (SD_estimator.uwave_r, SD_estimator.mvlue_r):
            raise ValueError(f'invalid std_dev method {std_dev}')
        return Xbar_statistic().sd(data, std_dev=std_dev, sizes=sizes)

    def limits(self, center, std_dev, sizes, conf):
        if conf < 0:
            raise ValueError(f'invalid conf argument {conf}')

        if all(size == sizes[0] for size in sizes):
            sizes = [sizes[0]]
        sizes = np.array(sizes)

        Rtab = len(_se_R_unscaled)

        if conf >= 1:
            if any(size > Rtab for size in sizes):
                raise ValueError(f'group size must be less than {Rtab + 1} when giving nsigmas')
            seR = std_dev * np.array([_se_R_unscaled[size] for size in sizes])
            lcl = center - conf * seR
            ucl = center + conf * seR
        else:
            # ucl <- qtukey(1 - (1 - conf)/2, sizes, 1e100) * std.dev
            # lcl <- qtukey((1 - conf)/2, sizes, 1e100) * std.dev
            raise NotImplementedError(
                'Tukey studentized range distribution not available from scipy. Use nsigmas instead')
        lcl[lcl < 0] = 0
        return pd.DataFrame({'LCL': lcl, 'UCL': ucl})


class S_statistic(Base_statistic):
    """ Statistics used in computing and drawing a Shewhart S chart """
    qcc_type = 'S'
    description = ('standard deviation', 'standard deviations of a continuous variable')

    def stats(self, data, sizes=None):
        if sizes is None:
            sizes = self.getSizes(data)
        by_group = np.nanstd(data, ddof=1, axis=1)
        return GroupMeans(by_group, sum(sizes * by_group) / sum(sizes))

    def sd(self, data, std_dev=None, sizes=None):
        if isinstance(std_dev, numbers.Number):
            return std_dev

        if std_dev is None:
            std_dev = SD_estimator.uwave_sd

        if std_dev not in (SD_estimator.uwave_sd, SD_estimator.mvlue_sd, SD_estimator.rmsdf):
            raise ValueError(f'invalid std_dev method {std_dev}')
        return Xbar_statistic().sd(data, std_dev=std_dev, sizes=sizes)

    def limits(self, center, std_dev, sizes, conf):
        if conf < 0:
            raise ValueError(f'invalid conf argument {conf}')

        if all(size == sizes[0] for size in sizes):
            sizes = [sizes[0]]
        sizes = np.array(sizes)

        c4 = np.array([qcc_c4(size) for size in sizes])
        se_stats = std_dev * np.sqrt(1 - c4**2)
        if conf >= 1:
            lcl = center - conf * se_stats
            ucl = center + conf * se_stats
        else:
            p = (1 - conf) / 2
            chi2 = [stats.chi2(size - 1).ppf(p) for size in sizes]
            lcl = std_dev * np.sqrt(chi2 / (sizes - 1))

            chi2 = [stats.chi2(size - 1).ppf(1 - p) for size in sizes]
            ucl = std_dev * np.sqrt(chi2 / (sizes - 1))
        lcl[lcl < 0] = 0
        return pd.DataFrame({'LCL': lcl, 'UCL': ucl})


class P_statistic(Base_statistic):
    """ Statistics used in computing and drawing a Shewhart p chart """
    qcc_type = 'p'
    description = ('proportion', 'proportion of nonconforming units')

    @staticmethod
    def getSizes(data):
        raise ValueError('p-charts require argument sizes to be provided')

    def stats(self, data, sizes=None):
        if sizes is None:
            raise ValueError('p-charts require argument sizes to be provided')
        if isinstance(sizes, numbers.Number):
            sizes = np.array([sizes] * len(data))
        elif isinstance(sizes, (pd.Series, pd.DataFrame)):
            sizes = sizes.values
        if isinstance(data, (pd.Series, pd.DataFrame)):
            data = data.values
        data = data.flatten()
        sizes = sizes.flatten()
        return GroupMeans(data / sizes, sum(data) / sum(sizes))

    def sd(self, data, std_dev=None, sizes=None):
        if isinstance(std_dev, numbers.Number):
            return std_dev
        if isinstance(sizes, numbers.Number):
            sizes = np.array([sizes] * len(data))
        else:
            sizes = np.array(sizes)
        if isinstance(data, (pd.Series, pd.DataFrame)):
            data = data.values
        pbar = np.sum(data) / np.sum(sizes)
        return np.sqrt(pbar * (1 - pbar))

    def limits(self, center, std_dev, sizes, conf):
        if all(size == sizes[0] for size in sizes):
            sizes = [sizes[0]]
        sizes = np.array(sizes)
        limits = NP_statistic().limits(center * sizes, std_dev, sizes, conf)

        if all(size == sizes[0] for size in sizes):
            sizes = [sizes[0]]
        sizes = np.array(sizes)

        limits['LCL'] = limits['LCL'] / sizes
        limits['UCL'] = limits['UCL'] / sizes
        return limits


class NP_statistic(Base_statistic):
    """ Statistics used in computing and drawing a Shewhart np chart """
    qcc_type = 'np'
    description = ('count', 'number of nonconforming units')

    def stats(self, data, sizes=None):
        if sizes is None:
            raise ValueError('np-charts require argument sizes to be provided')
        if isinstance(sizes, numbers.Number):
            sizes = np.array([sizes] * len(data))
        pbar = np.sum(data) / np.sum(sizes)
        center = sizes * pbar
        if all(c == center[0] for c in center):
            center = center[0]
        return GroupMeans(data, center)

    def sd(self, data, std_dev=None, sizes=None):
        if isinstance(std_dev, numbers.Number):
            return std_dev
        if isinstance(sizes, numbers.Number):
            sizes = np.array([sizes] * len(data))
        else:
            sizes = np.array(sizes)
        if isinstance(data, (pd.Series, pd.DataFrame)):
            data = data.values

        pbar = np.sum(data) / np.sum(sizes)
        std_dev = np.sqrt(sizes * pbar * (1 - pbar))
        if all(c == std_dev[0] for c in std_dev):
            std_dev = std_dev[0]
        return std_dev

    def limits(self, center, std_dev, sizes, conf):
        if conf < 0:
            raise ValueError(f'invalid conf argument {conf}')

        if all(size == sizes[0] for size in sizes):
            sizes = [sizes[0]]
        sizes = np.array(sizes)

        pbar = np.mean(center / sizes)
        if conf >= 1:
            tol = conf * np.sqrt(pbar * (1 - pbar) * sizes)
            lcl = center - tol
            ucl = center + tol
        else:
            p = (1 - conf) / 2
            lcl = np.array([stats.binom(size, p).ppf(pbar) for size in sizes])
            ucl = np.array([stats.binom(size, p).isf(pbar) for size in sizes])
        lcl[lcl < 0] = 0
        ucl[ucl > sizes] = sizes
        return pd.DataFrame({'LCL': lcl, 'UCL': ucl})

# # c Chart
#
# stats.c <- function(data, sizes)
# {
#   data <- as.vector(data)
#   sizes <- as.vector(sizes)
#   if (length(unique(sizes)) != 1)
#      stop("all sizes must be be equal for a c chart")
#   statistics <- data
#   center <- mean(statistics)
#   list(statistics = statistics, center = center)
# }
#
# sd.c <- function(data, sizes, ...)
# {
#   data <- as.vector(data)
#   std.dev <- sqrt(mean(data))
#   return(std.dev)
# }
#
# limits.c <- function(center, std.dev, sizes, conf)
# {
#   if (conf >= 1)
#      { lcl <- center - conf * sqrt(center)
#        lcl[lcl < 0] <- 0
#        ucl <- center + conf * sqrt(center)
#      }
#   else
#      { if (conf > 0 & conf < 1)
#           { ucl <- qpois(1 - (1 - conf)/2, center)
#             lcl <- qpois((1 - conf)/2, center)
#           }
#        else stop("invalid conf argument. See help.")
#      }
#   limits <- matrix(c(lcl, ucl), ncol = 2)
#   rownames(limits) <- rep("", length = nrow(limits))
#   colnames(limits) <- c("LCL", "UCL")
#   return(limits)
# }
#
# # u Chart
#
# stats.u <- function(data, sizes)
# {
#   data <- as.vector(data)
#   sizes <- as.vector(sizes)
#   statistics <- data/sizes
#   center <- sum(sizes * statistics)/sum(sizes)
#   list(statistics = statistics, center = center)
# }
#
# sd.u <- function(data, sizes, ...)
# {
#   data <- as.vector(data)
#   sizes <- as.vector(sizes)
#   std.dev <- sqrt(sum(data)/sum(sizes))
#   return(std.dev)
# }
#
# limits.u <- function(center, std.dev, sizes, conf)
# {
#   sizes <- as.vector(sizes)
#   if (length(unique(sizes))==1)
#      sizes <- sizes[1]
#   limits.c(center * sizes, std.dev, sizes, conf) / sizes
# }


def qcc_c4(n):
    return np.sqrt(2 / (n - 1)) * np.exp(gammaln(n / 2) - gammaln((n - 1) / 2))


qccStatistics = QCCStatistics()



'''
Created on Jul 9, 2020

@author: petergedeck
'''
from string import ascii_uppercase

from matplotlib.gridspec import GridSpec
from scipy import linalg, optimize
from scipy import stats

import matplotlib.patches as patches
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import statsmodels.formula.api as smf
import matplotlib.transforms as transforms


class MahalanobisT2:
    def __init__(self, x, factor_name, response_names=None, conf_level=0.95, compare_to=None):
        if response_names is None:
            response_names = [c for c in x.columns if c != factor_name]
        x = x[[factor_name, *response_names]]
        x = x.rename(columns={factor_name: 'factor'})
        D = x.groupby('factor').mean()
        self.groupmeans = D
        nfactors = len(D)
        # self.total_cov = x.cov()
        # self.total_cov_inv = linalg.inv(self.total_cov)
        cov = None
        for _, subset in x.groupby('factor'):
            if cov is None:
                cov = subset[response_names].cov()
            else:
                cov = cov + subset[response_names].cov()
        cov = cov / nfactors
        self.cov = cov

        N = np.unique(x.groupby('factor').apply(lambda x: len(x)))
        if len(N) > 1:
            print(f'different number of measures by factor, using n={N[0]}')
        N = N[0]
        p = D.shape[1]
        K = N**2 / (2 * N) * (2 * N - p - 1) / ((2 * N - 2) * p)
        Qf = stats.f(p, 2 * N - p - 1).ppf(conf_level)
        self.Qf = Qf
        self.K = K
        D2 = D.iloc[0] - D.iloc[1]
        self.D2 = D2
        D3 = pd.DataFrame([[0] * p, D.iloc[0] - D.iloc[1]], columns=list(ascii_uppercase[:p]))
        lm = smf.ols(formula=f"A ~ {'+'.join(D3.columns[1:])} - 1", data=D3).fit()

        def fun(x, cov_inv):
            x = pd.DataFrame(x, columns=list(ascii_uppercase[1:p]))
            D4 = np.array([lm.predict(x), *x.values]).flatten()
            delta = D4 - D2
            return abs(K * np.transpose(delta) @ cov_inv @ delta - Qf)
        cov_inv = linalg.inv(cov)
        opt_result = optimize.minimize(fun, [0] * (D3.shape[1] - 1), args=(cov_inv, ), method='Nelder-Mead')

        res = pd.DataFrame(opt_result.x, columns=list(ascii_uppercase[1:p]))
        res = np.array([lm.predict(res), *res.values]).flatten()
        res = np.array([res, D2.values, (2 * D2 - res).values])
        res = res[res[:, 1].argsort()[::-1]]
        res = pd.DataFrame(res, columns=response_names, index=('LCR', 'Center', 'UCR'))

        self.coord = res
        self.mahalanobis = res.apply(lambda x: np.sqrt(x @ cov_inv @ x), axis=1)
        self.mahalanobis_compare = None
        if compare_to is not None:
            self.mahalanobis_compare = np.sqrt(compare_to @ cov_inv @ compare_to)
#             x = pd.DataFrame([self.mahalanobis_compare], columns=list(ascii_uppercase[1:p]))
#             self.mahalanobis_compare_coord = [self.mahalanobis_compare, lm.predict(x)]
#
#             def fun(x, cov_inv):
#                 x = pd.DataFrame(x, columns=list(ascii_uppercase[1:p]))
#             opt_result = optimize.minimize(fun, [0] * (D3.shape[1] - 1), args=(cov_inv, ), method='Nelder-Mead')
#             res = pd.DataFrame(opt_result.x, columns=list(ascii_uppercase[1:p]))
#             res = np.array([lm.predict(res), *res.values]).flatten()
#             self.mahalanobis_compare_coord = res

    def summary(self):
        print('Coordinates')
        print(self.coord)
        print()
        print('Mahalanobis')
        print(self.mahalanobis)
        print()
        if self.mahalanobis_compare is not None:
            print(f' comparison: {self.mahalanobis_compare:.5g}')

    def plot(self):
        if len(self.coord.columns) > 2:
            raise ValueError('Plot only supported for two dimensions')
        fig = plt.figure(figsize=[4, 5])
        gs = GridSpec(2, 1, height_ratios=[3, 2])
        ax1 = fig.add_subplot(gs[0])
        ax2 = fig.add_subplot(gs[1])
        col1, col2 = self.coord.columns
        ax1.plot([0, self.coord[col1][0]], [0, self.coord[col2][0]], color='black')
        ax1.plot(0, 0, marker='o', markeredgecolor='black', markerfacecolor='white')
        ax1.plot(self.coord[col1], self.coord[col2], color='red',
                 marker='o', markeredgecolor='black', markerfacecolor='black')
        # Add confidence ellipse
        radius = np.sqrt(self.Qf / self.K)
        confidence_ellipse(self.D2.values, self.cov.values, ax=ax1, n_std=radius, edgecolor='grey')
        ax1.set_xlabel(f'difference {col1}')
        ax1.set_ylabel(f'difference {col2}')

        left = 0.5
        if self.mahalanobis_compare:
            left = 0.35
        ax2.set_xlim(0, 2)
        mmin = min(*self.mahalanobis, self.mahalanobis_compare)
        mmax = max(*self.mahalanobis, self.mahalanobis_compare)
        mdelta = mmax - mmin
        ax2.set_ylim(mmin - 0.15 * mdelta, mmax + 0.15 * mdelta)
        for direction in ('top', 'bottom', 'right'):
            ax2.spines[direction].set_visible(False)
        ax2.get_xaxis().set_visible(False)
        mmin = min(self.mahalanobis)
        mmax = max(self.mahalanobis)
        ax2.add_patch(patches.Rectangle((left, mmin), 1, mmax - mmin, facecolor="red", edgecolor="black"))
        y = self.mahalanobis[1]
        ax2.plot((left, left + 1), (y, y), color='black', linewidth=2)
        if self.mahalanobis_compare is not None:
            y = self.mahalanobis_compare
            ax2.plot((left - 0.1, left + 1.1), (y, y), linewidth=2)
            ax2.text(left + 1.15, y, 'Comparison', verticalalignment='center')
        ax2.set_ylabel(r'Mahalanobis $T^2$')

        plt.tight_layout()
        return ax1, ax2


def confidence_ellipse(means, cov, ax, n_std=3.0, facecolor='none', **kwargs):
    """
    Create a plot of the covariance confidence ellipse of *x* and *y*.

    Parameters
    ----------
    x, y : array-like, shape (n, )
        Input data.

    ax : matplotlib.axes.Axes
        The axes object to draw the ellipse into.

    n_std : float
        The number of standard deviations to determine the ellipse's radiuses.

    **kwargs
        Forwarded to `~matplotlib.patches.Ellipse`

    Returns
    -------
    matplotlib.patches.Ellipse
    """
    pearson = cov[0, 1] / np.sqrt(cov[0, 0] * cov[1, 1])
    # Using a special case to obtain the eigenvalues of this
    # two-dimensionl dataset.
    ell_radius_x = np.sqrt(1 + pearson)
    ell_radius_y = np.sqrt(1 - pearson)
    ellipse = patches.Ellipse((0, 0), width=ell_radius_x * 2, height=ell_radius_y * 2,
                              facecolor=facecolor, **kwargs)

    # Calculating the stdandard deviation of x from
    # the squareroot of the variance and multiplying
    # with the given number of standard deviations.
    scale_x = np.sqrt(cov[0, 0]) * n_std
    mean_x = means[0]

    # calculating the stdandard deviation of y ...
    scale_y = np.sqrt(cov[1, 1]) * n_std
    mean_y = means[1]

    transf = transforms.Affine2D() \
        .rotate_deg(45) \
        .scale(scale_x, scale_y) \
        .translate(mean_x, mean_y)

    ellipse.set_transform(transf + ax.transData)
    return ax.add_patch(ellipse)



'''
Created on Dec 30, 2020

@author: petergedeck
'''
import numpy as np

from scipy.linalg.special_matrices import toeplitz
from scipy.linalg import inv
from statsmodels.tsa.stattools import acf
from scipy import stats
from collections import namedtuple


def optimalLinearPredictor(x, n, nlags=10):
    ns = len(x)
    res = np.array(x)
    for i in range(n + 1, ns + 1):
        xs = x[:i]
        ak = acf(xs, nlags=nlags, fft=True)
        R = toeplitz(ak[:nlags])
        r = ak[1:]
        b = np.matmul(inv(R), r)
        res[i - 1] = np.sum(x[(i - 2):(i - nlags - 2):-1] * b)
    return res


def quadraticPredictor(x, n, s):
    ns = len(x)
    res = np.array(x)
    nx = np.arange(0, -n, -1)
    A = np.column_stack([[1] * n, nx, nx**2])
    invAA = inv(np.matmul(A.transpose(), A))
    invAAA = np.matmul(invAA, A.transpose())
    ts = np.array([1, s, s**2])
    for i in range(n, ns - s + 1):
        xs = x[i - n:i]
        xs = xs[::-1]
        b = np.matmul(invAAA, xs)
        pred = np.sum(b * ts)
        res[i + s - 1] = pred
    return res


def masPredictor(x, m, s):
    n = len(x)
    wm = np.arange(-m, m + 1)
    res = np.array(x)
    for i in range(m, n - m - s):
        xtm = x[i - m:i + m + 1]
        b0m = np.mean(xtm)
        b1m = 3 * np.sum(xtm * wm) / (m * (m + 1) * (2 * m + 1))
        res[i + m + s] = b0m + (m + s) * b1m
    return res


def normRandomWalk(n, v, w, c, seed=None):
    np.random.seed(seed)
    s1 = np.sqrt(v)
    s2 = np.sqrt(w)
    X = np.zeros(n)
    Tet = np.zeros(n)
    cf = np.zeros(n)
    Mt = np.zeros(n)
    cf[0] = (c + w) * v / (c + w + v)
    Tet[0] = s2 * stats.norm.rvs()
    X[0] = Tet[0] + s1 * stats.norm.rvs()
    for i in range(1, n):
        Tet[i] = Tet[i - 1] + s2 * stats.norm.rvs()
        X[i] = Tet[i] + s1 * stats.norm.rvs()
        cf[i] = (cf[i - 1] + w) * v / (cf[i - 1] + w + v)
        at = (cf[i] + w) / (cf[i] + w + v)
        Mt[i] = (1 - at) * Mt[i - 1] + at * X[i]
    nvrmResult = namedtuple('nvrmResult', 't,X,predicted')
    return nvrmResult(np.arange(n), X, Mt)


def dlmLinearGrowth(X, C0, v, W, M0):
    n = len(X)
    pred = np.ones(n)
    for i in range(n):
        at = np.array([1.0, i + 1]).transpose()
        pred[i] = np.matmul(at.transpose(), M0)
        ut = v + np.matmul(at.transpose(), np.matmul(C0 + W, at))
        ei = X[i] - np.matmul(at.transpose(), M0)
        ev = np.array([ei / ut, (i + 1) * ei / ut]).transpose()
        M0 = M0 + np.matmul((C0 + W), ev)
        B = np.array([[1 / ut, (i + 1) / ut], [(i + 1) / ut, (i + 1)**2 / ut]])
        C0 = (C0 + W) - np.matmul(C0 + W, np.matmul(B, (C0 + W)))
    return pred


def simulateARMA(n, a, b, seed=None):
    np.random.seed(seed)
    p = len(a)
    q = len(b)
    pm = max(p, q)
    ns = n + pm

    X = np.zeros(ns)
    X[0] = stats.norm.rvs()
    for i in range(1, pm):
        X[i] = X[i - 1] + stats.norm.rvs()
    E = stats.norm.rvs(size=ns)

    for i in range(pm, ns):
        X[i] = np.sum(X[i - p:i][::-1] * a)
        E[i] = np.sum(E[i - q:i][::-1] * b) + E[i]
    X = X + E
    return X


def predictARMA(X, a):
    n = len(X)
    p = len(a)
    pred = np.zeros(n)
    pred[:p] = X[:p]
    for i in range(p, n):
        pred[i] = np.sum(X[i - p:i][::-1] * a)
    return pred



'''
Created on Jun 22, 2020

@author: petergedeck
'''
#
# Functions used to signal points out of control

#
from itertools import groupby
from numbers import Number
import itertools

import numpy as np
import pandas as pd


RUN_LENGTH = 7


def shewhartRules(qcc, **kwargs):
    return {
        **beyondLimits(qcc, **kwargs),
        **violatingRuns(qcc, **kwargs),
    }


def beyondLimits(qcc, limits=None, **kwargs):
    limits = qcc.limits if limits is None else limits
    statistics = list(qcc.stats.statistics)
    if qcc.newstats:
        statistics.extend(qcc.newstats.statistics)
    if isinstance(statistics, pd.Series):
        statistics = statistics.values
    statistics = np.array(statistics).flatten()

    if len(limits['UCL']) == 1:
        return {'beyondLimits': {'UCL': np.nonzero(statistics > limits['UCL'][0])[0],
                                 'LCL': np.nonzero(statistics < limits['LCL'][0])[0]}}
    return {'beyondLimits': {'UCL': np.nonzero(statistics.ravel() > limits['UCL'].ravel())[0],
                             'LCL': np.nonzero(statistics.ravel() < limits['LCL'].ravel())[0]}}


def violatingRuns(qcc, run_length=7, **kwargs):
    if run_length == 0:
        return {'violatingRuns': []}
    center = qcc.center
    statistics = list(qcc.stats.statistics)
    if qcc.newstats:
        statistics.extend(qcc.newstats.statistics)
    if isinstance(statistics, pd.Series):
        statistics = statistics.values
    statistics = np.array(statistics).flatten()
    if isinstance(center, Number) or len(center.shape) == 0:
        diffs = statistics - center
    elif center.shape[0] == 1:
        diffs = statistics - center[0]
    else:
        raise NotImplementedError()
    diffs[diffs > 0] = 1
    diffs[diffs < 0] = -1
    runs = run_length_encoding(diffs, as_list=True)
    violating = runs >= run_length
    violating_above = violating & (diffs > 0)
    violating_below = violating & (diffs < 0)

    violators = []
    for violating in (violating_above, violating_below):
        current = 0
        for length, violates in run_length_encoding(violating):
            if violates:
                violators.extend(range(current + run_length - 1, current + length))
            current += length
    return {'violatingRuns': violators}


def run_length_encoding(sequence, as_list=False):
    rle = [(len(list(g)), k) for k, g in groupby(sequence)]
    if as_list:
        def flatten(x): return itertools.chain(*x)
        rle = np.array(list(flatten([count] * count for count, _ in rle)))
    return rle



'''
Created on Jun 20, 2020

@author: petergedeck
'''
from scipy import stats

from mistat.qcc.qualityControlChart import QualityControlChart
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd


class ProcessCapability:
    """ Computes process capability indices for a qcc object of type 'xbar' """

    def __init__(self, qcc, spec_limits, std_dev=None, target=None, nsigmas=None,
                 confidence_level=0.95):
        if not isinstance(qcc, QualityControlChart):
            raise ValueError('Argument qcc must be a QualityControlChart object')
        if qcc.qcc_type not in ('xbar', 'xbarone'):
            raise ValueError('Process Capability Analysis only available for charts type "xbar" and "xbar.one" charts')

        self.qcc = qcc
        data = self.qcc.data
        self.nobs = np.count_nonzero(~np.isnan(data))
        self.center = qcc.center
        self.std_dev = std_dev or qcc.std_dev

        lsl = spec_limits[0]
        usl = spec_limits[1]
        self.spec_limits = pd.DataFrame({'LSL': [lsl], 'USL': [usl]})

        self.has_target = target is not None
        self.target = target or np.nanmean(self.spec_limits)
        if target is None:
            self.target = np.nanmean(self.spec_limits)
            self.has_target = True

        # validation:
        # - only one of lcl or ucl can be NaN
        # - target must be between self.spec_limits

        self.nsigmas = nsigmas or self.qcc.nsigmas
        if not self.nsigmas:
            raise ValueError('nsigmas not available from qcc object. Provide nsigmas as argument.')

        self.confidence_level = confidence_level
        self.calcPCIndices()

        self.exp_LSL = np.NaN
        if not np.isnan(lsl):
            self.exp_LSL = stats.norm.cdf((lsl - self.center) / self.std_dev) * 100
        self.exp_USL = np.NaN
        if not np.isnan(usl):
            self.exp_USL = (1 - stats.norm.cdf((usl - self.center) / self.std_dev)) * 100
        self.obs_LSL = np.nanmean(data < lsl) * 100
        self.obs_USL = np.nanmean(data > usl) * 100

    def calcPCIndices(self):
        usl = self.spec_limits.values[0, 1]
        lsl = self.spec_limits.values[0, 0]
        nsigmas = self.nsigmas
        std_dev = self.std_dev
        center = self.center
        target = self.target

        self.Cp = (usl - lsl) / (2 * nsigmas * std_dev)
        self.Cp_u = (usl - self.center) / (nsigmas * std_dev)
        self.Cp_l = (self.center - lsl) / (nsigmas * std_dev)
        self.Cp_k = min(self.Cp_u, self.Cp_l)
        self.Cpm = self.Cp / np.sqrt(1 + ((center - target) / std_dev) ** 2)

        alpha = 1 - self.confidence_level
        alphas = [alpha / 2, 1 - alpha / 2]
        n = self.nobs
        self.Cp_limits = [self.Cp * np.sqrt(stats.chi2(n - 1).ppf(a) / (n - 1)) for a in alphas]
        self.Cp_l_limits = [self.Cp_l * (1 + f * stats.norm.ppf(self.confidence_level) *
                                         np.sqrt(1 / (9 * n * self.Cp_l**2) + 1 / (2 * (n - 1))))
                            for f in (-1, 1)]
        self.Cp_u_limits = [self.Cp_u * (1 + f * stats.norm.ppf(self.confidence_level) *
                                         np.sqrt(1 / (9 * n * self.Cp_u**2) + 1 / (2 * (n - 1))))
                            for f in (-1, 1)]
        self.Cp_k_limits = [self.Cp_k * (1 + f * stats.norm.ppf(1 - alpha / 2) *
                                         np.sqrt(1 / (9 * n * self.Cp_k**2) + 1 / (2 * (n - 1))))
                            for f in (-1, 1)]

        ratio = (center - target) / std_dev
        df = n * (1 + ratio**2) / (1 + 2 * ratio**2)
        self.Cpm_limits = [self.Cpm * np.sqrt(stats.chi2(df).ppf(a) / df) for a in alphas]

    def summary(self):
        print('Process Capability Analysis')
        print()
        print(f'Number of obs = {self.nobs:<4d}         Target = {self.target:.2f}')
        print(f'       Center = {self.center:<14.2f}  LSL = {self.spec_limits.values[0, 0]:.2f}')
        print(f'       StdDev = {self.std_dev:<14.6f}  USL = {self.spec_limits.values[0, 1]:.2f}')
        print()
        print('Capability indices:')
        print()
        print('        Value     2.5%   97.5%')
        for key in ('Cp', 'Cp_l', 'Cp_u', 'Cp_k', 'Cpm'):
            print(f'{key:<5s} {getattr(self, key):8.4f}', end='')
            limits = getattr(self, f'{key}_limits')
            print(f'{limits[0]:8.4f}{limits[1]:8.4f}')
        print()
        print(f'Exp<LSL {self.exp_LSL:3.0f}%   Obs<LSL {self.obs_LSL:3.0f}%')
        print(f'Exp>USL {self.exp_USL:3.0f}%   Obs>USL {self.obs_USL:3.0f}%')

    def plot(self, bins='scott', ax=None):
        if ax is None:
            _, ax = plt.subplots(figsize=(8, 6))
        data = self.qcc.data
        if isinstance(data, pd.DataFrame):
            data = data.values
        data = data.flatten()
        data = data[~np.isnan(data)]
        xlim = [min(np.min(data), np.min(self.spec_limits.values), self.target),
                max(np.max(data), np.max(self.spec_limits.values), self.target)]
        xlim = xlim + np.diff(xlim) * [-0.1, 0.1]

        ax.hist(data, bins=bins, density=True, color='lightgrey', edgecolor='grey')
        xx = np.linspace(*xlim, num=250)
        ax.plot(xx, stats.norm(self.center, self.std_dev).pdf(xx), ls='--', color='gray')

        position = [self.spec_limits.LSL[0], self.target, self.spec_limits.USL[0]]
        ax.axvline(position[0], color='red', ls=':')
        ax.axvline(position[1], color='red', ls='--')
        ax.axvline(position[2], color='red', ls=':')

        ax.secondary_xaxis('bottom')
        ax.xaxis.tick_top()
        ax.set_xticks(position)
        ax.set_xticklabels(['LSL', 'Target', 'USL'])
        ax.set_yticks([])

        fig = ax.get_figure()
        fig.subplots_adjust(bottom=0.25)
        options = {'fontsize': 11}
        col = [0.1, 0.35, 0.55, 0.75]
        row = [0.18, 0.14, 0.1, 0.06, 0.02]
        fig.text(col[0], row[0], f'Number of obs = {self.nobs}', **options)
        fig.text(col[0], row[1], f'Center = {self.center:.5g}', **options)
        fig.text(col[0], row[2], f'StdDev = {self.std_dev:.5g}', **options)

        fig.text(col[1], row[0], f'Target = {self.center:.2g}', **options)
        fig.text(col[1], row[1], f'LSL = {self.spec_limits.LSL[0]:.2g}', **options)
        fig.text(col[1], row[2], f'USL = {self.spec_limits.USL[0]:.2g}', **options)

        fig.text(col[2], row[0], f'Cp = {self.Cp:.2g}', **options)
        fig.text(col[2], row[1], f'Cp_l = {self.Cp_l:.2g}', **options)
        fig.text(col[2], row[2], f'Cp_u = {self.Cp_u:.2g}', **options)
        fig.text(col[2], row[3], f'Cp_k = {self.Cp_k:.2g}', **options)
        fig.text(col[2], row[4], f'Cpm = {self.Cpm:.2g}', **options)

        fig.text(col[3], row[0], f'Exp<LSL = {self.exp_LSL:.2g}%', **options)
        fig.text(col[3], row[1], f'Exp>USL = {self.exp_USL:.2g}%', **options)
        fig.text(col[3], row[2], f'Obs<LSL = {self.obs_LSL:.2g}%', **options)
        fig.text(col[3], row[3], f'Obs>USL = {self.obs_USL:.2g}%', **options)
        return ax


# process.capability <- function(object, spec.limits, target, std.dev, nsigmas, confidence.level = 0.95, breaks="scott", add.stats=TRUE, print=TRUE, digits = getOption("digits"), restore.par=TRUE)
# {
# # Computes process capability indices for a qcc object of type "xbar"
# # and plot the histogram

#   title <- paste("Process Capability Analysis\nfor", object$data.name)
#



from .cusum import Cusum, cusumArl, cusumPfaCed
from .ewmaChart import EWMA
from .paretoChart import ParetoChart
from .processCapability import ProcessCapability
from .qualityControlChart import qcc_groups, QualityControlChart
from .shro import shroArlPfaCedNorm
from .statistics import qccStatistics



'''
Created on Jun 28, 2020

@author: petergedeck
'''
from collections import namedtuple
import math

from scipy import stats
from scipy.optimize import root_scalar

from mistat.acceptanceSampling.distributions import OChypergeom, OCpoisson
from mistat.acceptanceSampling.oc import OC_TYPES
import numpy as np


RiskPoint = namedtuple('RiskPoint', 'pdefect,paccept')
Plan = namedtuple('Plan', 'n,c,r')
PlanNormal = namedtuple('PlanNormal', 'n,k,s_type')


def findPlan(PRP, CRP, oc_type='binomial', N=None, s_type='known'):
    """ Find the sampling plan with the smallest sample size, which meets a prespecified 
    Producer and Consumer Risk Points.

    The convention used here, as in many books, is to use equality for the Producer Risk 
    Point rather than the consumer risk point.

    No consideration is given to "cost functions
    """
    oc_type = oc_type.lower()
    if oc_type not in [*OC_TYPES, 'normal']:
        raise ValueError(f'Unknown type {oc_type}')
    PRP = checkRiskPoint(PRP, 'Producer risk point', oc_type)
    CRP = checkRiskPoint(CRP, 'Consumer risk point', oc_type)
    if np.any(PRP[0] >= CRP[0]):
        raise ValueError('Consumer Risk Point quality must be greater than Producer Risk Point quality')

    if oc_type == 'binomial':
        '''
    c <- 0
    n <- c+1
    repeat {
      if (calc.OCbinomial(n=n,c=c,r=c+1,pd=CRP[1]) > CRP[2])
        n <- n + 1
      else if (calc.OCbinomial(n=n,c=c,r=c+1,pd=PRP[1]) < PRP[2])
        c <- c + 1
      else
        break
    }
    return(list(n=n, c=c, r=c+1))
    '''

        raise NotImplementedError
    elif oc_type == 'hypergeom':
        if N is None:
            raise ValueError('N must be supplied for the hypergeometric distribution.')
        c = np.array([0])
        n = c + 1
        while True:
            if OChypergeom(n=n, c=c, r=c + 1, N=N, D=CRP[0] * N).paccept > CRP[1]:
                n += 1
            elif OChypergeom(n=n, c=c, r=c + 1, N=N, D=PRP[0] * N).paccept < PRP[1]:
                c += 1
            else:
                break
        return Plan(n[0], c[0], (c + 1)[0])

    elif oc_type == 'poisson':
        raise NotImplementedError
        c = 0
        n = c + 1
        while True:
            if OCpoisson(n=[n], c=[c], r=[c + 1], pd=CRP[0]).paccept > CRP[1]:
                print(n, c, OCpoisson(n=[n], c=[c], r=[c + 1], pd=CRP[0]).paccept)
                n += 1
            elif OCpoisson(n=[n], c=[c], r=[c + 1], pd=PRP[0]).paccept < PRP[1]:
                print(n, c, OCpoisson(n=[n], c=[c], r=[c + 1], pd=PRP[0]).paccept)
                c += 1
            else:
                break
            break
        return Plan(n, c, c + 1)

        '''
    c <- 0
    n <- c+1
    repeat {
      if (calc.OCpoisson(n=n,c=c,r=c+1,pd=CRP[1]) > CRP[2])
        n <- n + 1
      else if (calc.OCpoisson(n=n,c=c,r=c+1,pd=PRP[1]) < PRP[2])
        c <- c + 1
      else
        break
    }
    return(list(n=n, c=c, r=c+1))
  }
    '''
    elif oc_type == 'normal':
        s_type = s_type.lower()
        if s_type not in ['known', 'unknown']:
            raise ValueError(f'Unknown s_type {s_type}')
        # With known standard deviation
        if s_type == 'known':
            n = stats.norm.ppf(1 - PRP[1])[0] + stats.norm.ppf(CRP[1])[0]
            n = n / (stats.norm.ppf(CRP[0])[0] - stats.norm.ppf(PRP[0])[0])
            n = int(np.ceil(n * n))
            k = stats.norm.ppf(1 - PRP[1])[0] / np.sqrt(n) - stats.norm.ppf(PRP[0])[0]
            return PlanNormal(n, k, s_type)
        elif s_type == 'unknown':
            n = 2
            k = find_k(n, PRP[0], PRP[1], interval=[0, 1000])
            nc = - stats.norm.ppf(CRP[0])[0] * np.sqrt(n)
            pa = 1 - stats.nct.cdf(k * np.sqrt(n), df=n - 1, nc=nc)
            while pa > CRP[1]:
                n = n + 1
                k = find_k(n, PRP[0], PRP[1], interval=[0, 1000])
                nc = - stats.norm.ppf(CRP[0])[0] * np.sqrt(n)
                pa = 1 - stats.nct.cdf(k * np.sqrt(n), df=n - 1, nc=nc)
            return PlanNormal(n, k, s_type)


def findPlanApprox(PRP, CRP, N):
    """ Calculate single-stage sampling plan using approximation of hypergeometric distribution """
    alpha = (1 - PRP[1])
    beta = CRP[1]
    p0 = PRP[0]
    pt = CRP[0]

    za_pq0 = stats.norm.ppf(1 - alpha) * np.sqrt(p0 * (1 - p0))
    zb_pqt = stats.norm.ppf(1 - beta) * np.sqrt(pt * (1 - pt))

    # calculate approximation of n
    n0 = ((za_pq0 + zb_pqt) / (pt - p0)) ** 2
    nstar = int(round(n0 / (1 + n0 / N)))

    # calculate approximation of c
    cstar = int(round(nstar * p0 - 0.5 + za_pq0 * np.sqrt(nstar * (1 - nstar / N))))
    return Plan(nstar, cstar, cstar + 1)


def checkRiskPoint(RP, info, oc_type):
    if not isinstance(RP, list) and len(RP) != 2:
        raise ValueError(f'{info} must be a list with two elements (pdefect, paccept)')
    pdefect, paccept = RP
    if not isinstance(pdefect, list):
        pdefect = [pdefect]
    if not isinstance(paccept, list):
        paccept = [paccept]
    pdefect = np.array(pdefect)
    paccept = np.array(paccept)
    if oc_type == 'poisson':
        if not np.all(0 < pdefect):
            raise ValueError(f'{info}: all quality values of must be larger than 0')
        if not np.all(0 < paccept):
            raise ValueError(f'{info}: all paccept values of must be larger than 0')
    else:
        if not np.all(0 < pdefect < 1):
            raise ValueError(f'{info}: all values of pdefect must fall within [0, 1]')
        if not np.all(0 < paccept < 1):
            raise ValueError(f'{info}: all values of paccept must fall within [0, 1]')

    return RiskPoint(pdefect, paccept)


# Utility to find k for a given n


def find_k(n, pd, pa, interval=None):
    """ find k for a given n """
    if interval is None:
        interval = [0, 5]
    df = n - 1
    sqrt_n = np.sqrt(n)
    nc = - stats.norm.ppf(pd) * sqrt_n
    pa = 1 - pa

    def f(x):
        return stats.nct.cdf(x * sqrt_n, df=df, nc=nc) - pa
    return root_scalar(f, bracket=interval, method='brentq').root


"""
find.plan <- function(PRP, CRP,
                      type=c("binomial","hypergeom","poisson","normal"),
                      N,
                      s.type=c("known", "unknown"))
{
 
  else if(CRP[1] <= PRP[1])
    stop("Consumer Risk Point quality must be greater than Producer Risk Point quality")

  ## Attributes Sampling Plan - Binomial distribution
  if (type == "binomial") {
    c <- 0
    n <- c+1
    repeat {
      if (calc.OCbinomial(n=n,c=c,r=c+1,pd=CRP[1]) > CRP[2])
        n <- n + 1
      else if (calc.OCbinomial(n=n,c=c,r=c+1,pd=PRP[1]) < PRP[2])
        c <- c + 1
      else
        break
    }
    return(list(n=n, c=c, r=c+1))
  }
  ## Attributes Sampling Plan - Hypergeometric distribution
  if (type == "hypergeom") {
    c <- 0
    n <- c+1
    repeat {
      if (calc.OChypergeom(n=n,c=c,r=c+1,N=N,D=CRP[1]*N) > CRP[2])
        n <- n + 1
      else if (calc.OChypergeom(n=n,c=c,r=c+1,N=N,D=PRP[1]*N) < PRP[2])
        c <- c + 1
      else
        break
    }
    return(list(n=n, c=c, r=c+1))
  }
  ## Attributes Sampling Plan - Poisson distribution
  if (type == "poisson") {
    c <- 0
    n <- c+1
    repeat {
      if (calc.OCpoisson(n=n,c=c,r=c+1,pd=CRP[1]) > CRP[2])
        n <- n + 1
      else if (calc.OCpoisson(n=n,c=c,r=c+1,pd=PRP[1]) < PRP[2])
        c <- c + 1
      else
        break
    }
    return(list(n=n, c=c, r=c+1))
  }
  ## Variables Sampling Plan - Normal distribution
  else if (type=="normal") {
    ## With known standard deviation
    if (s.type=="known") {
      n <- ceiling( ((qnorm(1-PRP[2]) + qnorm(CRP[2]))/
                     (qnorm(CRP[1])-qnorm(PRP[1])) )^2)
      k <- qnorm(1-PRP[2])/sqrt(n) - qnorm(PRP[1])
      return(list(n=n, k=k, s.type=s.type))
    }
    ## With unknown standard deviation
    else if (s.type=="unknown") {
      n <- 2 ## Need a minimum of 1 degree of freedom (=n-1) for the NC t-dist
      k <- find.k(n, PRP[1], PRP[2], interval=c(0,1000))
      pa <- 1- pt(k*sqrt(n), df=n-1, ncp=-qnorm(CRP[1])*sqrt(n))
      while(pa > CRP[2]){
        n <- n+1
        k <- find.k(n, PRP[1], PRP[2])
        pa <- 1-pt(k*sqrt(n), df=n-1, ncp=-qnorm(CRP[1])*sqrt(n))
      }
      return(list(n=n, k=k, s.type=s.type))
    }
  }
}
"""


def normalApproximationOfH(j, M, n, N):
    P = N / M
    Q = 1 - P
    return stats.norm.cdf((j + 0.5 - n * P) / math.sqrt(n * P * Q * (1 - n / M)))



'''

'''
from contextlib import redirect_stdout
from pathlib import Path
import io
import unittest

import pytest

from mistat.qcc.processCapability import ProcessCapability
from mistat.qcc.qualityControlChart import qcc_groups, QualityControlChart
from mistat.qcc.statistics import SD_estimator, QCCStatistics
from mistat.qcc.tests.utilities import assertQCC
import numpy as np
import pandas as pd


class TestProcessCapability(unittest.TestCase):
    def test_ProcessCapability(self):
        pistonrings = pd.read_csv(Path(__file__).parent / 'data' / 'pistonrings.csv')
        diameter = qcc_groups(pistonrings['diameter'], pistonrings['sample'])
        data = diameter[:25, ]

        qcc = QualityControlChart(data, qcc_type='xbar')
        assertQCC(qcc, 74.00118, 0.009785039, [73.98805, 74.0143])

        pc = ProcessCapability(qcc, [73.95, 74.05])
        assertPC(pc, 125, 74.00118, 0.009785039, [73.95, 74.05], 3)
        expected = {
            'Cp': (1.703, 1.491, 1.915),
            'Cp_l': (1.743, 1.555, 1.932),
            'Cp_u': (1.663,  1.483, 1.844),
            'Cp_k': (1.663, 1.448, 1.878),
            'Cpm': (1.691, 1.480, 1.902),
        }
        assertPCindices(pc, expected)

        pc = ProcessCapability(qcc, [73.95, 74.05], target=74.02)
        assertPC(pc, 125, 74.00118, 0.009785039, [73.95, 74.05], 3, target=74.02)
        expected = {
            'Cp': (1.703, 1.491, 1.915),
            'Cp_l': (1.743, 1.555, 1.932),
            'Cp_u': (1.663,  1.483, 1.844),
            'Cp_k': (1.663, 1.448, 1.878),
            'Cpm': (0.7856, 0.6556, 0.9154),
        }
        assertPCindices(pc, expected)
        assert pc.exp_LSL == pytest.approx(0, abs=1e-4)
        assert pc.exp_USL == pytest.approx(0, abs=1e-4)
        assert pc.obs_LSL == pytest.approx(0, abs=1e-4)
        assert pc.obs_USL == pytest.approx(0, abs=1e-4)

        pc = ProcessCapability(qcc, [73.99, 74.01])
        assertPC(pc, 125, 74.00118, 0.009785039, [73.99, 74.01], 3)
        expected = {
            'Cp': (0.3407, 0.2983, 0.3830),
            'Cp_l': (0.3807, 0.3176, 0.4439),
            'Cp_u': (0.3006, 0.2424, 0.3588),
            'Cp_k': (0.3006, 0.2312, 0.3700),
            'Cpm': (0.3382, 0.2960, 0.3804),
        }
        assertPCindices(pc, expected)
        assert pc.exp_LSL == pytest.approx(13, abs=0.5)
        assert pc.exp_USL == pytest.approx(18, abs=0.5)
        assert pc.obs_LSL == pytest.approx(12, abs=0.5)
        assert pc.obs_USL == pytest.approx(16, abs=0.5)

        pc = ProcessCapability(qcc, [73.99, 74.1])
        assertPC(pc, 125, 74.00118, 0.009785039, [73.99, 74.1], 3)
        expected = {
            'Cp': (1.8736,  1.6406, 2.1063),
            'Cp_l': (0.3807, 0.3176,  0.4439),
            'Cp_u': (3.3665, 3.0115,  3.7215),
            'Cp_k': (0.3807,  0.3055,  0.4559),
            'Cpm': (0.4083,  0.3377, 0.4788),
        }
        assertPCindices(pc, expected)
        assert pc.exp_LSL == pytest.approx(13, abs=0.5)
        assert pc.exp_USL == pytest.approx(0, abs=0.5)
        assert pc.obs_LSL == pytest.approx(12, abs=0.5)
        assert pc.obs_USL == pytest.approx(0, abs=0.5)

    def test_ProcessCapability_summary(self):
        pistonrings = pd.read_csv(Path(__file__).parent / 'data' / 'pistonrings.csv')
        diameter = qcc_groups(pistonrings['diameter'], pistonrings['sample'])
        data = diameter[:25, ]

        qcc = QualityControlChart(data, qcc_type='xbar')
        assertQCC(qcc, 74.00118, 0.009785039, [73.98805, 74.0143])

        pc = ProcessCapability(qcc, [73.99, 74.01])
        f = io.StringIO()
        with redirect_stdout(f):
            pc.summary()
        s = f.getvalue()
        assert 'Exp<LSL' in s
        assert '18%' in s
        assert 'Capability indices:' in s
        assert 'Target =' in s
        assert '74.00' in s


def assertPC(pc, nobs, center, std_dev, limits, nsigmas, target=None):
    assert pc.nobs == nobs
    assert pc.center == pytest.approx(center)
    assert pc.std_dev == pytest.approx(std_dev)
    np.testing.assert_allclose(pc.spec_limits, [limits], rtol=1e-5)
    assert pc.target == (target or np.mean(limits))
    assert pc.nsigmas == nsigmas


def assertPCindices(pc, expected):
    for key in expected:
        assert getattr(pc, key) == pytest.approx(expected[key][0], rel=5e-4), key
        k = f'{key}_limits'
        np.testing.assert_allclose(getattr(pc, k),
                                   expected[key][1:], rtol=5e-4, err_msg=k)



'''
Utility functions for "Data Mining for Business Analytics: Concepts, Techniques, and 
Applications in Python"

(c) 2019 Galit Shmueli, Peter C. Bruce, Peter Gedeck 
'''
from pathlib import Path
import unittest

import pytest

from mistat.qcc.qualityControlChart import qcc_groups, QualityControlChart
from mistat.qcc.statistics import SD_estimator, QCCStatistics
from mistat.qcc.tests.utilities import assertQCC, assertQCCviolations
import numpy as np
import pandas as pd


class TestComparisonWithR(unittest.TestCase):
    def test_pistonrings(self):
        pistonrings = pd.read_csv(Path(__file__).parent / 'data' / 'pistonrings.csv')
        diameter = qcc_groups(pistonrings['diameter'], pistonrings['sample'])
        data = diameter[:25, ]
        newdata = diameter[25:, ]

        qcc = QualityControlChart(data, qcc_type='xbar')
        assertQCC(qcc, 74.00118, 0.009785039, [73.98805, 74.0143])
        assertQCCviolations(qcc)
#         qcc(diameter[1:25,], type="xbar", newdata=diameter[26:40,])
#         assertNewQCC(qcc, 74.00118, 0.009785039, [73.98805, 74.0143])
        qcc = QualityControlChart(data, qcc_type='xbar', nsigmas=2)
        assertQCC(qcc, 74.00118, 0.009785039, [73.99242, 74.00993])
        assertQCCviolations(qcc, beyondLimits={'UCL': [0], 'LCL': [13]})
#         qcc(diameter[1:25,], type="xbar", newdata=diameter[26:40,])
#         assertNewQCC(qcc, 74.00118, 0.009785039, [73.98805, 74.0143])

        qcc = QualityControlChart(data, qcc_type='R')
        assertQCC(qcc, 0.02276, 0.009785039, [0, 0.04812533])

        qcc = QualityControlChart(data, qcc_type='S')
        assertQCC(qcc, 0.009240037, 0.009829977, [0, 0.019302416])

    def test_pistonrings_variable_control_limits(self):
        """
        # variable control limits
        out <- c(9, 10, 30, 35, 45, 64, 65, 74, 75, 85, 99, 100)
        diameter <- qcc.groups(pistonrings$diameter[-out], sample[-out])
        r <- qcc(diameter[1:25,], type="xbar")
        r$center
        r$std.dev
        r$limits
        r <- qcc(diameter[1:25,], type="R")
        qcc(diameter[1:25,], type="S")
        qcc(diameter[1:25,], type="xbar", newdata=diameter[26:40,])
        qcc(diameter[1:25,], type="R", newdata=diameter[26:40,])
        qcc(diameter[1:25,], type="S", newdata=diameter[26:40,])        
        """
        out = [idx - 1 for idx in [9, 10, 30, 35, 45, 64, 65, 74, 75, 85, 99, 100]]
        pistonrings = pd.read_csv(Path(__file__).parent / 'data' / 'pistonrings.csv')
        pistonrings = pistonrings.drop(pistonrings.index[out])
        diameter = qcc_groups(pistonrings['diameter'], pistonrings['sample'])
        data = diameter[:25, ]
        newdata = diameter[25:, ]

        qcc = QualityControlChart(data, qcc_type='xbar')
        assertQCC(qcc, 74.00075, 0.009856701, [[73.98753, 74.01398], [73.98368, 74.01782], [73.98753, 74.01398]])
        assertQCCviolations(qcc)
        qcc = QualityControlChart(data, qcc_type='xbar', nsigmas=2)
        assertQCC(qcc, 74.00075, 0.009856701, [[73.99194, 74.00957], [73.98937, 74.01213], [73.99194, 74.00957]])
        assertQCCviolations(qcc, beyondLimits={'UCL': [0], 'LCL': [13]})

        qcc = QualityControlChart(data, qcc_type='R')
        assertQCC(qcc, 0.02230088, 0.009856701, [[0, 0.04785198], [0, 0.04857007], [0, 0.04785198]])

        qcc = QualityControlChart(data, qcc_type='S')
        assertQCC(qcc, 0.00938731, 0.009930923, [[0, 0.01955302], [0, 0.02318885], [0, 0.01955302]])

    def test_orangejuice(self):
        """
        data(orangejuice)
        attach(orangejuice)
        r <- qcc(D[trial], sizes=size[trial], type="p")

        # remove out-of-control points (see help(orangejuice) for the reasons)
        inc <- setdiff(which(trial), c(15,23))
        q1 <- qcc(D[inc], sizes=size[inc], type="p")

        r <- qcc(D[inc], sizes=size[inc], type="p", newdata=D[!trial], newsizes=size[!trial])
        r$center
        r$std.dev
        r$limits
        detach(orangejuice)
        """
        oj = pd.read_csv(Path(__file__).parent / 'data' / 'orangejuice.csv')
        data = oj[oj['trial']]
        newdata = oj[~oj['trial']]

        qcc = QualityControlChart(data['D'], sizes=data['size'], qcc_type='p')
        assertQCC(qcc, 0.2313333, 0.421685, [0.05242755, 0.4102391])
        assertQCCviolations(qcc, beyondLimits={'UCL': [14, 22], 'LCL': []})

        dataRed = data.drop(data.index[qcc.violations['beyondLimits']['UCL']])
        qcc = QualityControlChart(dataRed['D'], sizes=dataRed['size'], qcc_type='p')
        assertQCC(qcc, 0.215, 0.4108223, [0.04070284, 0.3892972])
        assertQCCviolations(qcc, beyondLimits={'UCL': [19], 'LCL': []})

#         qcc = QualityControlChart(dataRed['D'], newdata=newdata, sizes=dataRed['size'], qcc_type='p')
#         assertQCC(qcc, 0.215, 0.4108223, [0.04070284, 0.3892972])

    def test_orangejuice2(self):
        """
        data(orangejuice2)
        attach(orangejuice2)
        names(D) <- sample
        r = qcc(D[trial], sizes=size[trial], type="p")
        q2 <- qcc(D[trial], sizes=size[trial], type="p", newdata=D[!trial], newsizes=size[!trial])
        detach(orangejuice2)
        """
        oj = pd.read_csv(Path(__file__).parent / 'data' / 'orangejuice2.csv')
        data = oj[oj['trial']]
        newdata = oj[~oj['trial']]

        qcc = QualityControlChart(data['D'], sizes=data['size'], qcc_type='p')
        assertQCC(qcc, 0.1108333, 0.3139256, [0, 0.2440207])
        assertQCCviolations(qcc)

#         qcc = QualityControlChart(dataRed['D'], newdata=newdata, sizes=dataRed['size'], qcc_type='p')
#         assertQCC(qcc, 0.215, 0.4108223, [0.04070284, 0.3892972])

    def test_xbar_one(self):
        # viscosity data (Montgomery, pag. 242)
        x = [33.75, 33.05, 34, 33.81, 33.46, 34.02, 33.68, 33.27, 33.49, 33.20, 33.62, 33.00, 33.54, 33.12, 33.84]
        qcc = QualityControlChart(x, qcc_type="xbarone")
        assertQCC(qcc, 33.52333, 0.4261651, [32.24484, 34.80183])

        qcc = QualityControlChart(x, qcc_type="xbarone", std_dev=SD_estimator.sd)
        assertQCC(qcc, 33.52333, 0.3415928, [32.49855, 34.54811])

        #  Water content of antifreeze data (Wetherill and Brown, 1991, p. 120)
        x = [2.23, 2.53, 2.62, 2.63, 2.58, 2.44, 2.49, 2.34, 2.95, 2.54, 2.60, 2.45, 2.17, 2.58, 2.57, 2.44, 2.38,
             2.23, 2.23, 2.54, 2.66, 2.84, 2.81, 2.39, 2.56, 2.70, 3.00, 2.81, 2.77, 2.89, 2.54, 2.98, 2.35, 2.53]
        qcc = QualityControlChart(x, qcc_type="xbarone")
        assertQCC(qcc, 2.569706, 0.1794541, [2.031344, 3.108068])
        assertQCCviolations(qcc)

        qcc = QualityControlChart(x, qcc_type="xbarone", std_dev='SD')
        assertQCC(qcc, 2.569706, 0.2216795, [1.904667, 3.234744])
        assertQCCviolations(qcc)

        expected = [0.1794541, 0.1897519, 0.1955224, 0.2019203, 0.2022154, 0.2041948, 0.2072357, 0.2121212, 0.2126056, 0.2143082, 0.2172579,
                    0.2189612, 0.2183180, 0.2178859, 0.2181558, 0.2172365, 0.2168714, 0.2151667, 0.2138331, 0.2128866, 0.2122988, 0.2108173,
                    0.2095927]
        XbarOne = QCCStatistics().get('xbarone')
        np.testing.assert_allclose([XbarOne.sd(x, k=k) for k in range(2, 25)], expected, rtol=1e-5)



'''
Utility functions for "Data Mining for Business Analytics: Concepts, Techniques, and 
Applications in Python"

(c) 2019 Galit Shmueli, Peter C. Bruce, Peter Gedeck 
'''
import unittest

from scipy.special import gammaln
import pytest

from mistat.data import load_data
from mistat.qcc import statistics
from mistat.qcc.statistics import QCCStatistics, SD_estimator, qcc_c4
import numpy as np


class Test_qccStatistics(unittest.TestCase):
    array_1 = np.array([[1, 2, 3], [1, 2, np.NaN]])
    array_2 = np.array([[1, 2, 3], [1, 2, 7]])

    def setUp(self):
        self.qccStatistics = QCCStatistics()
        unittest.TestCase.setUp(self)

    def test_default_statistic(self):
        assert self.qccStatistics.default == 'xbar'

        for statistic in self.qccStatistics:
            assert hasattr(statistic, 'description'), statistic.qcc_type

    def test_SD_estimator(self):
        assert SD_estimator.get(None, SD_estimator.mr) == SD_estimator.mr
        assert SD_estimator.get(SD_estimator.sd, SD_estimator.mr) == SD_estimator.sd
        assert SD_estimator.get('sd', SD_estimator.mr) == SD_estimator.sd
        assert SD_estimator.get('SD', SD_estimator.mr) == SD_estimator.sd
        with pytest.raises(ValueError):
            SD_estimator.get(123, SD_estimator.mr)
        with pytest.raises(ValueError):
            SD_estimator.get('unknown estimator', SD_estimator.mr)

    def test_Xbar_statistic(self):
        xbar = self.qccStatistics.get('xbar')
        assert isinstance(xbar, statistics.Xbar_statistic)
        assert xbar.qcc_type == 'xbar'

        assert xbar.getSizes(self.array_1) == [3, 2]
        stats = xbar.stats(self.array_1)
        np.testing.assert_array_equal(stats.statistics, [2, 1.5])
        assert stats.center == 1.8

        assert xbar.sd(self.array_1, 12345) == 12345
        assert xbar.sd(self.array_1) == pytest.approx(1.03393)
        assert xbar.sd(self.array_1, std_dev=SD_estimator.uwave_r) == pytest.approx(1.03393)
        assert xbar.sd(self.array_1, std_dev=SD_estimator.uwave_sd) == pytest.approx(1.007303)
        assert xbar.sd(self.array_1, std_dev=SD_estimator.mvlue_r) == pytest.approx(1.085444)
        assert xbar.sd(self.array_1, std_dev=SD_estimator.mvlue_sd) == pytest.approx(1.049987)
        assert xbar.sd(self.array_1, std_dev=SD_estimator.rmsdf) == pytest.approx(0.9908318)

        sd = xbar.sd(self.array_1)
        sizes = xbar.getSizes(self.array_1)
        conf_limits = xbar.limits(stats.center, sd, sizes, 2.0)
        np.testing.assert_allclose(conf_limits, np.array([[0.6061206, 2.993879], [0.3378024, 3.262198]]), rtol=1e-4)

        conf_limits = xbar.limits(stats.center, sd, sizes, 0.95)
        np.testing.assert_allclose(conf_limits, np.array([[0.6300197, 2.969980], [0.3670726, 3.232927]]), rtol=1e-4)

        sizes = [3, 3]
        conf_limits = xbar.limits(stats.center, sd, sizes, 2.0)
        np.testing.assert_allclose(conf_limits, np.array([[0.6061206, 2.993879]]), rtol=1e-4)

        conf_limits = xbar.limits(stats.center, sd, sizes, 0.95)
        np.testing.assert_allclose(conf_limits, np.array([[0.6300197, 2.969980]]), rtol=1e-4)

    def test_S_statistic(self):
        S = self.qccStatistics.get('S')
        assert isinstance(S, statistics.S_statistic)
        assert S.qcc_type == 'S'

        assert S.getSizes(self.array_1) == [3, 2]
        stats = S.stats(self.array_1)
        assert stats.center == pytest.approx(0.8828427)
        np.testing.assert_allclose(stats.statistics, [1.0000000, 0.7071068], rtol=1e-5)

        assert S.sd(self.array_1, 12345) == 12345
        assert S.sd(self.array_1) == pytest.approx(1.007303)
        assert S.sd(self.array_1, std_dev=SD_estimator.uwave_sd) == pytest.approx(1.007303)
        assert S.sd(self.array_1, std_dev=SD_estimator.mvlue_sd) == pytest.approx(1.049987)
        assert S.sd(self.array_1, std_dev=SD_estimator.rmsdf) == pytest.approx(0.9908318)

        sd = S.sd(self.array_1)
        sizes = S.getSizes(self.array_1)
        conf_limits = S.limits(stats.center, sd, sizes, 3.0)
        np.testing.assert_allclose(conf_limits, np.array([[0, 2.28274627669134], [0, 2.70448059144019]]), rtol=1e-4)
        conf_limits = S.limits(stats.center, sd, sizes, 0.9)
        np.testing.assert_allclose(conf_limits, np.array(
            [[0.228134225169997, 1.74345862934255], [0.063164728444235, 1.97427769221479]]), rtol=1e-4)

        stats = S.stats(self.array_2)
        sd = S.sd(self.array_2)
        sizes = S.getSizes(self.array_2)
        conf_limits = S.limits(stats.center, sd, sizes, 3.0)
        np.testing.assert_allclose(conf_limits, np.array([[0, 5.41184]]), rtol=1e-4)

        conf_limits = S.limits(stats.center, sd, sizes, 0.9)
        np.testing.assert_allclose(conf_limits, np.array([[0.5385259, 4.115549]]), rtol=1e-4)

    def test_P_statistic(self):
        p = self.qccStatistics.get('p')
        assert isinstance(p, statistics.P_statistic)
        assert p.qcc_type == 'p'

        data = load_data('JANDEFECT')
        sizes = 100
        with pytest.raises(ValueError):
            p.stats(data)
        stats = p.stats(data, sizes)
        assert stats.center == pytest.approx(0.05387097)
        assert stats.statistics[0] == 0.06

        assert p.sd(data, 12345) == 12345
        assert p.sd(data, sizes=100) == pytest.approx(0.2257629)
        assert p.sd(data, sizes=[100] * len(data)) == pytest.approx(0.2257629)

        sd = p.sd(data, sizes=100)
        sizes = [100] * len(data)
        conf_limits = p.limits(stats.center, sd, sizes, 3.0)
        np.testing.assert_allclose(conf_limits, np.array([[0, 0.1215998]]), rtol=1e-4)

        conf_limits = p.limits(stats.center, sd, sizes, 0.9)
        np.testing.assert_allclose(conf_limits, np.array([[0.02, 0.09]]), rtol=1e-4)

    def test_NP_statistic(self):
        NP = self.qccStatistics.get('np')
        assert isinstance(NP, statistics.NP_statistic)
        assert NP.qcc_type == 'np'

        data = load_data('JANDEFECT')
        sizes = 100
        with pytest.raises(ValueError):
            NP.stats(data)
        stats = NP.stats(data, sizes)
        assert stats.center == pytest.approx(5.387097)
        assert stats.statistics[0] == 6

        assert NP.sd(data, 12345) == 12345
        assert NP.sd(data, sizes=100) == pytest.approx(2.257629)
        assert NP.sd(data, sizes=[100] * len(data)) == pytest.approx(2.257629)
#
        sd = NP.sd(data, sizes=100)
        sizes = [100] * len(data)
        conf_limits = NP.limits(stats.center, sd, sizes, 3.0)
        np.testing.assert_allclose(conf_limits, np.array([[0, 12.15998]]), rtol=1e-4)

        conf_limits = NP.limits(stats.center, sd, sizes, 0.9)
        np.testing.assert_allclose(conf_limits, np.array([[2, 9]]), rtol=1e-4)

    def test_R_statistic(self):
        R = self.qccStatistics.get('R')
        assert isinstance(R, statistics.R_statistic)
        assert R.qcc_type == 'R'

        data = load_data('CONTACTLEN').values
        assert R.getSizes(data) == [5] * 20

        stats = R.stats(data)
        assert stats.center == pytest.approx(0.23665)
        assert stats.statistics[0] == 0.119

        assert R.sd(data, 12345) == 12345
        assert R.sd(data) == pytest.approx(0.1017412)

        sd = R.sd(data)
        sizes = R.getSizes(data)
        conf_limits = R.limits(stats.center, sd, sizes, 3.0)
        np.testing.assert_allclose(conf_limits, np.array([[0, 0.5003893]]), rtol=1e-4)

        # conf_limits = R.limits(stats.center, sd, sizes, 0.9)
        # np.testing.assert_allclose(conf_limits, np.array([[0.1047873, 0.3924825]]), rtol=1e-4)

    def test_Xbar_one_statistic(self):
        data = [2.23, 2.53, 2.62, 2.63, 2.58, 2.44, 2.49, 2.34, 2.95, 2.54, 2.60, 2.45, 2.17, 2.58, 2.57, 2.44,
                2.38, 2.23, 2.23, 2.54, 2.66, 2.84, 2.81, 2.39, 2.56, 2.70, 3.00, 2.81, 2.77, 2.89, 2.54, 2.98, 2.35, 2.53]

        XbarOne = self.qccStatistics.get('xbarone')
        assert isinstance(XbarOne, statistics.Xbar_one_statistic)
        assert XbarOne.qcc_type == 'xbarone'

        stats = XbarOne.stats(data)
        assert stats.center == pytest.approx(2.569706)
        assert stats.statistics[0] == 2.23

        assert XbarOne.sd(data, 12345) == 12345
        assert XbarOne.sd(data) == pytest.approx(0.1794541)
        assert XbarOne.sd(data, std_dev=SD_estimator.mr) == pytest.approx(0.1794541)
        assert XbarOne.sd(data, std_dev=SD_estimator.sd) == pytest.approx(0.2216795)

        sd = XbarOne.sd(data)
        conf_limits = XbarOne.limits(stats.center, sd, None, 3.0)
        np.testing.assert_allclose(conf_limits, np.array([[2.031344, 3.108068]]), rtol=1e-4)

        conf_limits = XbarOne.limits(stats.center, sd, None, 0.9)
        np.testing.assert_allclose(conf_limits, np.array([[2.27453, 2.864882]]), rtol=1e-4)

    def test_qcc_c4(self):
        assert gammaln(2) == 0
        assert gammaln(3) == pytest.approx(0.6931472)
        assert qcc_c4(2) == pytest.approx(0.7978846)
        assert qcc_c4(3) == pytest.approx(0.8862269)
        assert qcc_c4(4) == pytest.approx(0.9213177)

    def test_abc(self):
        import mistat
        Ps = mistat.PistonSimulator(n_simulation=5 * 20, seed=1).simulate()
        Ps = mistat.simulationGroup(Ps, 5)

        cycleTime = mistat.qcc_groups(Ps['seconds'], Ps['group'])
        mistat.QualityControlChart(cycleTime)



'''
Created on May 3, 2021

@author: petergedeck
'''
import numpy as np
import pandas as pd
from scipy.linalg import eig


class ResponseSurfaceMethod:
    ''' Class provides information about the response surface defined by a 
    second order regression model

     See section on "Canonical representation" in book for mathematical details'''

    def __init__(self, model, codes):
        ''' determine vector b and array B from model '''
        b = np.array([model.params.get(c, 0) for c in codes])
        B = np.zeros([len(codes), len(codes)])
        for i, code_i in enumerate(codes):
            B[i, i] = model.params.get(f'np.power({code_i}, 2)', 0)
            for j, code_j in enumerate(codes[i + 1:], i + 1):
                B[i, j] = model.params.get(f'{code_i}:{code_j}', 0) / 2
                B[j, i] = B[i, j]
        self.model = model
        self.codes = codes
        self.b = b
        self.B = B

    def f(self, x):
        ''' calculate function value at position x '''
        exog = pd.Series(x, index=self.codes)
        return self.model.predict(exog=exog)[0]

    def gradient(self, x):
        ''' calculate gradient at position x '''
        gradient = pd.Series(self.b + np.matmul(self.B, x), index=self.codes)
        return gradient / np.linalg.norm(gradient)

    def stationary_point(self):
        return pd.Series(- 0.5 * np.matmul(np.linalg.inv(self.B), self.b), index=self.codes)

    def _eigen(self):
        ev, evec = eig(self.B)
        return {'eval': ev, 'evec': evec}

    def eigen(self):
        ev, evec = eig(self.B)
        indx = ev.real.argsort()[::-1]
        return {'eval': ev.real[indx], 'evec': evec[:, indx]}



'''
Created on Jul 9, 2020

@author: petergedeck
'''
import unittest

import pytest

from mistat.qcc.shro import shroArlPfaCedNorm, runLengthShroNorm, runLengthShroPois,\
    shroArlPfaCedPois
import numpy as np


class TestCusum(unittest.TestCase):
    def test_shroArlPfaCedNorm(self):
        x = [-1.0080707,  1.3549394, -0.4689749,  1.4681936,  0.4425564, 0.1462031,  0.1715031,
             0.5925072,  2.7647493, 0.6192188, 0.3945324, -1.6215868, -0.9459741, -0.9978776]

        res = runLengthShroNorm(x, 1, 1, 10, 1, 19)
        assert res['rl'] == 9
        assert len(res['w']) == 9
        assert res['w'][0] is None
        for e1, e2 in zip(res['w'][1:], [2.344282e-01, 3.469896e-09, 7.275562e-01, 4.415837e-05, 1.319939e-06, 1.699855e-06, 1.145011e-04, 3.110184e+05]):
            assert e1 == pytest.approx(e2)

        res = runLengthShroPois(x, 1.001, 1, 0.58)
        assert res['rl'] == 7
        assert len(res['w']) == 7
        assert res['w'][0] is None
        for e1, e2 in zip(res['w'][1:], [0.3683780, 0.5031622, 0.5537946, 0.5718620, 0.5783402, 0.5807385]):
            assert e1 == pytest.approx(e2)

        res = runLengthShroPois(x, 1.001, 0.9, 0.58)
        assert res['rl'] == 4
        assert len(res['w']) == 4
        assert res['w'][0] is None
        for e1, e2 in zip(res['w'][1:], [0.4071206, 0.5718245, 0.6399946]):
            assert e1 == pytest.approx(e2)

        res = runLengthShroPois(x, 11 / 6, 1, 19)
        assert res['rl'] == np.inf
        assert len(res['w']) == 14
        assert res['w'][0] is None
        for e1, e2 in zip(res['w'][1:], [0.8363378, 0.5083987, 1.3511733, 1.1310696, 0.8566233, 0.7578374, 0.9260974, 3.7859976, 2.5626072,  1.6646799, 0.3668443, 0.2834033, 0.25786117291]):
            assert e1 == pytest.approx(e2)

        res = shroArlPfaCedNorm(seed=1, N=10, verbose=False)['statistic']
        assert res['ARL'] == pytest.approx(135.0)
        assert res['Std. Error'] == pytest.approx(40.54873)

        res = shroArlPfaCedNorm(seed=1, N=10, mean0=10, tau=10, verbose=False)['statistic']
        assert res['ARL'] == pytest.approx(10.7)
        assert res['Std. Error'] == pytest.approx(0.51088159)
        assert res['PFA'] == pytest.approx(0.1)
        assert res['CED'] == pytest.approx(1.22222222)
        assert res['CED-Std. Error'] == pytest.approx(3.72107039)

    def test_shroArlPfaCedPois(self):
        x = [12, 8, 15, 13, 11, 10, 8, 10, 17, 16,
             89, 85, 120, 98, 109, 110, 93, 85, 98, 109]
        res = runLengthShroPois(x, 1.1, 1, 19)
        assert res['rl'] == 10

        x = [7, 7, 6, 22, 16, 11, 11, 16, 9, 12,
             112, 121, 95, 98, 114, 103, 99, 97, 86, 105]
        res = runLengthShroPois(x, 1.1, 1, 19)
        assert res['rl'] == 8

        res = shroArlPfaCedPois(seed=1, verbose=False)['statistic']
        assert res['ARL'] == pytest.approx(24.79)
        assert res['Std. Error'] == pytest.approx(1.570178)

        res = shroArlPfaCedPois(seed=1, tau=10, lambda1=100, verbose=False)['statistic']
        assert res['ARL'] == pytest.approx(10.89)
        assert res['Std. Error'] == pytest.approx(0.044486, abs=0.0001)
        assert res['PFA'] == pytest.approx(0.03)
        assert res['CED'] == pytest.approx(0.958763)
        assert res['CED-Std. Error'] == pytest.approx(1.108611)



'''
Created on Jul 7, 2020

@author: petergedeck
'''
from mistat.qcc.statistics import Base_statistic, qccStatistics
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd


def ewmaSmooth(y, x=None, smooth=0.20, start=None):
    """
    Exponential-Weighted Moving Average 

    Return smooth values based on 

    z_t = smooth*y_t + (1-smooth)*z_t-1      

    where 0<= smooth <=1 is the parameter which controls the weights applied 
    to the data, and start is the starting value.

    Returns a list with elements:
    x = ordered x-values
    y = smoothed fitted values of y

    if x is provided, y values will be reordered 
    """
    if not (0 <= abs(smooth) <= 1):
        raise ValueError('smooth parameter must be between 0 and 1')
    if x is None:
        x = list(range(len(y)))
    else:
        if len(x) != len(y):
            raise ValueError('x and y must have the same length')
        x, y = zip(*sorted(zip(x, y)))

    last = start if start is not None else y[0]
    z = []
    for yi in y:
        last = smooth * yi + (1 - smooth) * last
        z.append(last)
    return {
        'x': np.array(x),
        'y': np.array(z),
        'smooth': smooth,
        'start': start,
    }


class EWMA:
    def __init__(self, data, sizes=None, center=None, std_dev=None, smooth=0.2, nsigmas=3, newdata=None):
        if sizes is None:
            sizes = Base_statistic.getSizes(data)
        elif isinstance(sizes, int):
            sizes = [sizes] * len(data)
        else:
            if len(sizes) != len(data):
                raise ValueError('sizes and data must have the same length')
        qcc_type = 'xbarone' if set(sizes) == {1} else 'xbar'

        qccStatistic = qccStatistics.get(qcc_type)
        statistics = qccStatistic.stats(data, sizes)

        center = center
        if center is None:
            center = statistics.center
        std_dev = qccStatistic.sd(data, std_dev, sizes=sizes)

        self.data = data
        self.statistics = statistics.statistics
        self.sizes = np.array(sizes)
        self.center = center
        self.std_dev = std_dev

        if newdata is not None:
            raise NotImplementedError('requires implementation')

        self.ewma = ewmaSmooth(self.statistics, smooth=smooth, start=center)
        self.x = self.ewma['x']
        self.y = self.ewma['y']
        self.smooth = smooth

        L1 = smooth / (2 - smooth)
        f = 2 * np.arange(1, len(self.statistics) + 1)
        L2 = 1 - (1 - smooth) ** f

        sigma2 = std_dev ** 2 / self.sizes * L1 * L2
        self.sigma = np.sqrt(sigma2)
        self.nsigmas = nsigmas
        self.ucl = center + nsigmas * np.sqrt(sigma2)
        self.lcl = center - nsigmas * np.sqrt(sigma2)
        self.violations = list(np.where(self.y < self.lcl)[0])
        self.violations.extend(np.where(self.y > self.ucl)[0])
        self.violations = np.array(sorted(self.violations))

    def plot(self, ax=None):
        if ax is None:
            _, ax = plt.subplots(figsize=(7, 6))

        pd.Series(self.statistics).plot(style='.', marker='+', color='black')
        v = np.array(self.violations)
        ax.plot(self.x, self.y, color='grey')
        ax.plot(self.x[v], self.y[v], linestyle='None', marker='s', color='red')
        ax.plot(np.delete(self.x, v), np.delete(self.y, v), linestyle='None', marker='o', color='black', markersize=4)
        ax.axhline(self.center, color='grey')
        ax.plot(self.x, self.ucl, color='grey', linestyle='--')
        ax.plot(self.x, self.lcl, color='grey', linestyle='--')

        ax2 = ax.twinx()
        ax2.set_ylim(*ax.get_ylim())
        ax2.yaxis.tick_right()
        ax2.set_yticks([self.lcl[-1], self.ucl[-1], self.center])
        ax2.set_yticklabels(['LCL', 'UCL', 'CL'])

        ax.set_title('EWMA Chart')
        ax.set_xlabel('Group')
        ax.set_ylabel('Group summary statistics')

        fig = ax.get_figure()
        fig.subplots_adjust(bottom=0.23)
        fig.text(0.1, 0.1, f'Number of groups = {len(self.statistics)}', fontsize=12)
        fig.text(0.1, 0.06, f'Center = {self.center:.5g}', fontsize=12)
        fig.text(0.1, 0.02, f'StdDev = {self.std_dev:.5g}', fontsize=12)

        fig.text(0.5, 0.1, f'Smoothing parameter = {self.smooth}', fontsize=12)
        fig.text(0.5, 0.06, f'Control limits at {self.nsigmas}*sigma', fontsize=12)
        fig.text(0.5, 0.02, f'No. of points beyond limits = {len(self.violations)}', fontsize=12)

        return ax



'''
Created on Jun 30, 2020

@author: petergedeck
'''
from scipy import stats

import numpy as np
import pandas as pd

from .dodge_base import AcceptanceSamplingPlan


#' Single sampling plans for the binomial, hypergeometric and Poisson
#' distributions.
def SSPlanBinomial(N, n, Ac, p=None):
    p = p if p is not None else np.arange(0, 0.3, 0.001)
    p = np.array(p)

    OC = np.array([stats.binom(n, pi).cdf(Ac) for pi in p])

    AOQ = (N - n) * p * OC / N
    ATI = n * OC + N * (1 - OC)
    return AcceptanceSamplingPlan(p=p, OC=OC, n=[n] * len(p), AOQ=AOQ, ATI=ATI, ASN=None)


def SSPlanHyper(N, n, Ac, p=None):
    p = p if p is not None else np.arange(0, 0.3, 0.001)
    p = np.array(p)

    OC = np.array([stats.hypergeom(N, round(n * pi), n).cdf(Ac) for pi in p])

    AOQ = (N - n) * p * OC / N
    ATI = n * OC + N * (1 - OC)
    return AcceptanceSamplingPlan(p=p, OC=OC, n=[n] * len(p), AOQ=AOQ, ATI=ATI, ASN=None)


def SSPlanPoisson(N, n, Ac, p=None):
    p = p if p is not None else np.arange(0, 0.3, 0.001)
    p = np.array(p)

    OC = np.array([stats.poisson(n * pi).cdf(Ac) for pi in p])

    AOQ = (N - n) * p * OC / N
    ATI = n * OC + N * (1 - OC)
    return AcceptanceSamplingPlan(p=p, OC=OC, n=[n] * len(p), AOQ=AOQ, ATI=ATI, ASN=None)


#' Single Sampling Plan Designs
#'
#' Design a single sampling plan for given AQL, alpha, LQL, and beta. Currently
#' there are functions for the binomial and Poisson distributions.
#'
#' @param AQL Acceptable quality level
#' @param alpha producer's risk
#' @param LQL Limiting quality level
#' @param beta consumers' risk

def SSPDesignBinomial(AQL, alpha, LQL, beta):
    def nl(Ac, LQL, beta):
        n = 1
        while stats.binom(n, LQL).cdf(Ac) >= beta:
            n += 1
        return n

    def nu(Ac, AQL, alpha):
        n = 1
        while stats.binom(n, AQL).cdf(Ac) >= 1 - alpha:
            n += 1
        return n

    Ac = 0
    while nl(Ac, LQL, beta) > nu(Ac, AQL, alpha):
        Ac += 1
    n = nl(Ac, LQL, beta)
    return pd.Series({'n': n, 'Ac': Ac})


def SSPDesignPoisson(AQL, alpha, LQL, beta):
    def nl(Ac, LQL, beta):
        n = 1
        while stats.poisson(n * LQL).cdf(Ac) >= beta:
            n += 1
        return n

    def nu(Ac, AQL, alpha):
        n = 1
        while stats.poisson(n * AQL).cdf(Ac) >= 1 - alpha:
            n += 1
        return n

    Ac = 0
    while nl(Ac, LQL, beta) > nu(Ac, AQL, alpha):
        Ac += 1
    n = nl(Ac, LQL, beta)
    return pd.Series({'n': n, 'Ac': Ac})



'''
Created on Jun 16, 2020

@author: petergedeck
'''
from dataclasses import dataclass, field
from enum import Enum
from itertools import product
from typing import List

from scipy.stats import binom, hypergeom, poisson

import numpy as np
import pandas as pd


class OCtype(Enum):
    binomial = 'binomial'
    hypergeom = 'hypergeom'
    poisson = 'poisson'


def getDistribution(n, c, distribution, r=None, pd=None, N=None, D=None):
    if not isinstance(n, list):
        n = [n]
    if not isinstance(c, list):
        c = [c]
    if distribution == OCtype.binomial:
        return OCbinomial(n=n, c=c, r=r, pd=pd)
    elif distribution == OCtype.hypergeom:
        return OChypergeom(n=n, c=c, r=r, N=N, pd=pd, D=D)
    elif distribution == OCtype.poisson:
        return OCpoisson(n=n, c=c, r=r, pd=pd)
    else:
        raise NotImplementedError(f'Distribution {distribution} not implemented')


@dataclass
class OCdistribution:
    n: List[int]
    c: List[int]
    r: List[int] = field(default_factory=list)

    distribution: OCtype = field(init=False)
    paccept: List[float] = field(default_factory=list, init=False)

    def __post_init__(self):
        if self.r is None:
            if len(self.c) <= 2:
                self.r = [1 + self.c[-1]] * len(self.c)
            else:
                self.r = None

        self.c = np.array(self.c)
        self.r = np.array(self.r)


@dataclass
class OCbinomial(OCdistribution):
    pd: List[float] = field(default_factory=list)

    def __post_init__(self):
        super().__post_init__()
        self.distribution = OCtype.binomial

        if self.pd is None:
            self.pd = np.linspace(0, 1, 101)

        self.paccept = np.array([self.calcBinomial(pdi, self.n, self.c, self.r) for pdi in self.pd])

    def calcBinomial(self, p_d, n, c, r):
        # For each stage, find out all the possibilities which could
        # lead to still not having made a decision and then calculate
        # the appropriate probabilities.
        p_acc = 0.0
        for k in range(len(n)):
            # Determine combinations
            comb = [list(range(c[i] + 1, r[i])) for i in range(k)]
            x = pd.DataFrame(list(product(*comb)))
            # Calculate change from previous
            x.iloc[:, 1:] = x.iloc[:, 1:].values - x.iloc[:, :-1].values

            x[k] = c[k] - np.sum(x, axis=1)
            for _, xi in x.iterrows():
                p_acc += self.probAcc(xi.values, n, p_d)
        return p_acc

    @staticmethod
    def probAcc(x, n, p):
        k = len(x) - 1
        f = binom.cdf(x[k], n[k], p)
        for i in range(k):
            f *= binom.pmf(x[i], n[i], p)
        return f


@dataclass
class OChypergeom(OCdistribution):
    N: int = None
    pd: List[float] = field(default_factory=list)
    D: List[int] = field(default_factory=list)

    def __post_init__(self):
        super().__post_init__()
        self.distribution = OCtype.hypergeom
        if not self.D:
            self.D = [round(pdi * self.N) for pdi in self.pd]

        self.paccept = np.array([self.calcHypergeom(Di, self.n, self.c, self.r, self.N) for Di in self.D])

    def calcHypergeom(self, D, n, c, r, N):
        # For each stage, find out all the possibilities which could
        # lead to still not having made a decision and then calculate
        # the appropriate probabilities.
        p_acc = 0.0
        for k in range(len(c)):
            # Determine combinations
            comb = [list(range(c[i] + 1, r[i])) for i in range(k)]
            x = pd.DataFrame(list(product(*comb)))
            # Calculate change from previous
            x.iloc[:, 1:] = x.iloc[:, 1:].values - x.iloc[:, :-1].values

            x[k] = c[k] - np.sum(x, axis=1)
            for _, xi in x.iterrows():
                p_acc += self.probAcc(xi.values, n, N, D)
        return p_acc

    @staticmethod
    def probAcc(x, n, N, D):
        k = len(x)
        k1 = k - 1
        x_cumsum = np.cumsum(x)[0:k]
        n_cumsum = np.cumsum(n)
        D_cum = D - np.array([0, *x_cumsum[0:k1]])
        N_cum = N - np.array([0, *n_cumsum[0:k1]])
        f = hypergeom(N_cum[-1], max(0, D_cum[-1]), n[k1]).cdf(x[-1])
        for i in range(len(x) - 1):
            f *= hypergeom(N_cum[i], max(0, D_cum[i]), n[i]).pmf(x[i])
        return f


@dataclass
class OCpoisson(OCdistribution):
    pd: List[float] = field(default_factory=list)

    def __post_init__(self):
        super().__post_init__()
        self.distribution = OCtype.poisson

        if self.pd is None:
            self.pd = np.linspace(0, 1, 101)

        self.paccept = np.array([self.calcPoisson(pdi, self.n, self.c, self.r) for pdi in self.pd])

    def calcPoisson(self, p_d, n, c, r):
        # For each stage, find out all the possibilities which could
        # lead to still not having made a decision and then calculate
        # the appropriate probabilities.
        p_acc = 0.0
        for k in range(len(n)):
            # Determine combinations
            comb = [list(range(c[i] + 1, r[i])) for i in range(k)]
            x = pd.DataFrame(list(product(*comb)))
            # Calculate change from previous
            x.iloc[:, 1:] = x.iloc[:, 1:].values - x.iloc[:, :-1].values

            x[k] = c[k] - np.sum(x, axis=1)
            for _, xi in x.iterrows():
                p_acc += self.probAcc(xi.values, n, p_d)
                print(x, p_acc)
        return p_acc

    @staticmethod
    def probAcc(x, n, p):
        k = len(x) - 1
        f = poisson.cdf(x[k], np.floor(n[k] * p + 0.5))
        for i in range(k):
            f *= poisson.pmf(x[i], n[i] * p)
        return f



import gzip
from pathlib import Path

DEMO_FILE_DIR = Path(__file__).parent


def _readFile(filename):
    with gzip.open(filename) as f:
        return f.read()


def globalWarmingBlogs():
    articles = {
        'blog-1': _readFile(DEMO_FILE_DIR / 'global-warming-blog-1.txt.gz'),
        'blog-2': _readFile(DEMO_FILE_DIR / 'global-warming-blog-2.txt.gz'),
    }


def covid19Blogs():
    return {
        'article-1': _readFile(DEMO_FILE_DIR / 'covid-19-article-1.txt.gz'),
        'article-2': _readFile(DEMO_FILE_DIR / 'covid-19-article-2.txt.gz'),
        'blog-1': _readFile(DEMO_FILE_DIR / 'covid-19-blog-1.txt.gz'),
        'blog-2': _readFile(DEMO_FILE_DIR / 'covid-19-blog-2.txt.gz'),
        'blog-3': _readFile(DEMO_FILE_DIR / 'covid-19-blog-3.txt.gz'),
    }



'''
Utility functions for "Data Mining for Business Analytics: Concepts, Techniques, and 
Applications in Python"

(c) 2019 Galit Shmueli, Peter C. Bruce, Peter Gedeck 
'''
from pathlib import Path

import setuptools


def getVersion():
    f = Path(__file__).parent / 'src' / 'mistat' / 'version.py'
    lines = f.read_text().split('\n')
    version = [s for s in lines if '__version__' in s][0]
    version = version.split('=')[1].strip().strip("'")
    return version


with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="mistat",
    version=getVersion(),
    author="Peter Gedeck",
    author_email="mail@petergedeck.com",
    description="Utility functions for 'Modern Industrial Statistics'",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/gedeck/mistat",
    packages=setuptools.find_packages("src"),
    package_dir={'': 'src'},
    package_data={
        "mistat": ["csvFiles/*.csv.gz", "nlp/*.txt.gz"],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],

    test_suite='nose.collector',
    tests_require=['nose'],
)



'''
Created on Jun 22, 2020

@author: petergedeck
'''
from pathlib import Path
import unittest

from mistat.data import load_data
from mistat.qcc.qualityControlChart import QualityControlChart, qcc_groups
from mistat.qcc.rules import run_length_encoding
import numpy as np
import pandas as pd


class TestRules(unittest.TestCase):
    def test_shewhartRules(self):
        #         qcc = QualityControlChart(load_data('JANDEFECT'), qcc_type='np', sizes=100)
        #         print(qcc.violations)

        out = [idx - 1 for idx in [9, 10, 30, 35, 45, 64, 65, 74, 75, 85, 99, 100]]
        pistonrings = pd.read_csv(Path(__file__).parent / 'data' / 'pistonrings.csv')
        pistonrings = pistonrings.drop(pistonrings.index[out])
        diameter = qcc_groups(pistonrings['diameter'], pistonrings['sample'])
        qcc = QualityControlChart(diameter, qcc_type='xbar', center=74, std_dev=0.0098)

    def test_run_length_encoding(self):
        assert run_length_encoding('abbccc') == [(1, 'a'), (2, 'b'), (3, 'c')]
        assert run_length_encoding('abbcccaaaaa') == [(1, 'a'), (2, 'b'), (3, 'c'), (5, 'a')]
        assert run_length_encoding([]) == []
        np.testing.assert_array_equal(run_length_encoding('abbccc', as_list=True), [1, 2, 2, 3, 3, 3])



from .data import load_data, describe_data
from .design import mainEffectsPlot, interactionPlot, FDS_Plot
from .design import getModelMatrix, calculateMainEffects, calculateInteractions
from .ecdf import plotECDF
from .ml import plot_dendrogram
from .mqcc import MultivariateQualityControlChart, MahalanobisT2
from .qcc import Cusum, cusumArl, cusumPfaCed, shroArlPfaCedNorm
from .qcc import EWMA
from .qcc import qcc_groups, QualityControlChart, ParetoChart, ProcessCapability, qccStatistics
from .randomizationTest import randomizationTest
from .regression import stepwise_regression
from .reliability import availabilityEBD, renewalEBD
from .rsm import ResponseSurfaceMethod
from .runsTest import runsTest
from .simulation import PistonSimulator, SimulationResult, simulationGroup, PowerCircuitSimulation
from .stem_leaf import stemLeafDiagram
from .timeseries import dlmLinearGrowth, simulateARMA, predictARMA
from .timeseries import optimalLinearPredictor, quadraticPredictor, masPredictor, normRandomWalk



'''
Created on Jul 9, 2020

@author: petergedeck
'''
from collections import namedtuple

from scipy import linalg
from scipy import stats

from mistat.qcc.statistics import QCCStatistics
import numpy as np
import pandas as pd


GroupStatistics = namedtuple('GroupStatistics', 'statistics,means,center,cov')
DataSizes = namedtuple('DataSizes', 'num_samples,samples_sizes,num_variables')


class MQCCStatistics(QCCStatistics):

    def __init__(self):
        self.default = None
        self.statistics = {}
        self.register(T2_statistic, default=True)
        self.register(T2single_statistic)


class T2_statistic:
    """ Statistics used in computing and drawing a Shewhart xbar chart """
    qcc_type = 't2'
    description = ('T2', 'Hotelling T^2 chart for subgrouped data')

    def get_sizes(self, data):
        shapes = {k: v.shape for k, v in data.items()}
        num_samples = [v[0] for v in shapes.values()]  # m
        if len(set(num_samples)) == 1:
            num_samples = num_samples[0]
        sample_sizes = [v[1] for v in shapes.values()]  # n
        if len(set(sample_sizes)) == 1:
            sample_sizes = sample_sizes[0]
        p = len(data)  # number of variables
        return DataSizes(num_samples, sample_sizes, p)

    def stats(self, data, center=None, cov=None):
        num_samples, sample_sizes, p = self.get_sizes(data)

        # within sample means
        means = pd.DataFrame({k: np.nanmean(v, axis=1) for k, v in data.items()})
        # overall mean
        if center is None:
            center = pd.DataFrame([{k: np.nanmean(v) for k, v in data.items()}])
        else:
            center = pd.DataFrame(center)
        x = means.sub(center.values, axis=1)

        if cov is None:
            cov = np.zeros([p, p])
            for i in range(num_samples):
                sampleMean = means.iloc[i, :].values
                d = np.transpose(np.array([v[i] for k, v in data.items()]))
                d = d - sampleMean
                d = np.transpose(d) @ d
                cov += d / (sample_sizes - 1)
            cov = cov / num_samples
        cov_inv = linalg.inv(cov)

        T2 = sample_sizes * x.apply(lambda x: x @ cov_inv @ x, axis=1)
        return GroupStatistics(statistics=T2, means=means, center=center, cov=cov)

    def limits(self, ngroups, size, nvars, conf):
        m = ngroups  # num. of samples
        n = size  # samples sizes
        p = nvars  # num. of variables

        upper = p * (n - 1) / (m * n - m - p + 1) * stats.f(p, m * n - m - p + 1).ppf(conf)

        # Phase 1 control limits
        ucl = (m - 1) * upper
        lcl = 0
        control = pd.DataFrame([{'LCL': lcl, 'UCL': ucl}])

        # Phase 2 prediction limits
        ucl = (m + 1) * upper
        lcl = 0
        prediction = pd.DataFrame([{'LPL': lcl, 'UPL': ucl}])

        return {'control': control, 'prediction': prediction}


class T2single_statistic:
    """ Statistics used in computing and drawing a Shewhart xbar chart """
    qcc_type = 't2single'
    description = ('T2 single', 'Hotelling T^2 chart for individual observations')

    def get_sizes(self, data):
        return DataSizes(data.shape[0], 1, data.shape[1])

    def stats(self, data, center=None, cov=None):
        data = pd.DataFrame(data)
        m, _, _ = self.get_sizes(data)

        if center is None:
            center = np.mean(data, axis=0)
        else:
            center = pd.DataFrame(center)
        x = data.sub(center.values.flatten(), axis=1)
        if cov is None:
            cov = np.transpose(x) @ x
            cov = cov / (m - 1)
        cov_inv = linalg.inv(cov)
        T2 = x.apply(lambda x: x @ cov_inv @ x, axis=1)
        return GroupStatistics(statistics=T2, means=data, center=center, cov=cov)

    def limits(self, ngroups, size, nvars, conf):
        m = ngroups  # num. of samples
        n = size  # samples sizes # @UnusedVariable
        p = nvars  # num. of variables

        # Phase 1 control limits
        ucl = (m - 1)**2 / m * stats.beta(p / 2, (m - p - 1) / 2).ppf(conf)
        lcl = 0
        control = pd.DataFrame([{'LCL': lcl, 'UCL': ucl}])

        # Phase 2 prediction limits
        ucl = p * (m + 1) * (m - 1) / (m * (m - p)) * stats.f(p, m - p).ppf(conf)
        lcl = 0
        prediction = pd.DataFrame([{'LPL': lcl, 'UPL': ucl}])
        return {'control': control, 'prediction': prediction}


"""
> limits.T2.single
function (ngroups, size = 1, nvars, conf) 
{
    m <- ngroups
    n <- size
    p <- nvars
    ucl <- (m - 1)^2/m * qbeta(conf, p/2, (m - p - 1)/2)
    lcl <- 0
    ctrl.limits <- matrix(c(lcl, ucl), ncol = 2)
    ucl <- p * (m + 1) * (m - 1)/(m * (m - p)) * qf(conf, p, 
        m - p)
    lcl <- 0
    pred.limits <- matrix(c(lcl, ucl), ncol = 2)
    rownames(ctrl.limits) <- rownames(pred.limits) <- rep("", 
        nrow(pred.limits))
    colnames(ctrl.limits) <- c("LCL", "UCL")
    colnames(pred.limits) <- c("LPL", "UPL")
    return(list(control = ctrl.limits, prediction = pred.limits))
}
"""
mqccStatistics = MQCCStatistics()
